from __future__ import absolute_import,print_function
_a='\x07'
_Z='\ufeff'
_Y='\ue000'
_X='\ud7ff'
_W='\x85'
_V='%%%02X'
_U='version'
_T="-;/?:@&=+$,_.~*'()[]"
_S='---'
_R=' \n\x85\u2028\u2029'
_Q='\xa0'
_P='a'
_O='0'
_N=','
_M='...'
_L='\\'
_K='['
_J="'"
_I='?'
_H='"'
_G='!'
_F='\n\x85\u2028\u2029'
_E='\n'
_D=' '
_C=False
_B=None
_A=True
import sys
from .error import YAMLError,YAMLStreamError
from .events import *
from .compat import utf8,text_type,PY2,nprint,dbg,DBG_EVENT,check_anchorname_char
if _C:from typing import Any,Dict,List,Union,Text,Tuple,Optional;from .compat import StreamType
__all__=['Emitter','EmitterError']
class EmitterError(YAMLError):0
class ScalarAnalysis:
	def __init__(self,scalar,empty,multiline,allow_flow_plain,allow_block_plain,allow_single_quoted,allow_double_quoted,allow_block):self.scalar=scalar;self.empty=empty;self.multiline=multiline;self.allow_flow_plain=allow_flow_plain;self.allow_block_plain=allow_block_plain;self.allow_single_quoted=allow_single_quoted;self.allow_double_quoted=allow_double_quoted;self.allow_block=allow_block
class Indents:
	def __init__(self):self.values=[]
	def append(self,val,seq):self.values.append((val,seq))
	def pop(self):return self.values.pop()[0]
	def last_seq(self):
		try:return self.values[-2][1]
		except IndexError:return _C
	def seq_flow_align(self,seq_indent,column):
		if len(self.values)<2 or not self.values[-1][1]:return 0
		base=self.values[-1][0]if self.values[-1][0]is not _B else 0;return base+seq_indent-column-1
	def __len__(self):return len(self.values)
class Emitter:
	DEFAULT_TAG_PREFIXES={_G:_G,'tag:yaml.org,2002:':'!!'};MAX_SIMPLE_KEY_LENGTH=128
	def __init__(self,stream,canonical=_B,indent=_B,width=_B,allow_unicode=_B,line_break=_B,block_seq_indent=_B,top_level_colon_align=_B,prefix_colon=_B,brace_single_entry_mapping_in_flow_sequence=_B,dumper=_B):
		self.dumper=dumper
		if self.dumper is not _B and getattr(self.dumper,'_emitter',_B)is _B:self.dumper._emitter=self
		self.stream=stream;self.encoding=_B;self.allow_space_break=_B;self.states=[];self.state=self.expect_stream_start;self.events=[];self.event=_B;self.indents=Indents();self.indent=_B;self.flow_context=[];self.root_context=_C;self.sequence_context=_C;self.mapping_context=_C;self.simple_key_context=_C;self.line=0;self.column=0;self.whitespace=_A;self.indention=_A;self.compact_seq_seq=_A;self.compact_seq_map=_A;self.no_newline=_B;self.open_ended=_C;self.colon=':';self.prefixed_colon=self.colon if prefix_colon is _B else prefix_colon+self.colon;self.brace_single_entry_mapping_in_flow_sequence=brace_single_entry_mapping_in_flow_sequence;self.canonical=canonical;self.allow_unicode=allow_unicode;self.unicode_supplementary=sys.maxunicode>65535;self.sequence_dash_offset=block_seq_indent if block_seq_indent else 0;self.top_level_colon_align=top_level_colon_align;self.best_sequence_indent=2;self.requested_indent=indent
		if indent and 1<indent<10:self.best_sequence_indent=indent
		self.best_map_indent=self.best_sequence_indent;self.best_width=80
		if width and width>self.best_sequence_indent*2:self.best_width=width
		self.best_line_break=_E
		if line_break in['\r',_E,'\r\n']:self.best_line_break=line_break
		self.tag_prefixes=_B;self.prepared_anchor=_B;self.prepared_tag=_B;self.analysis=_B;self.style=_B;self.scalar_after_indicator=_A
	@property
	def stream(self):
		try:return self._stream
		except AttributeError:raise YAMLStreamError('output stream needs to specified')
	@stream.setter
	def stream(self,val):
		if val is _B:return
		if not hasattr(val,'write'):raise YAMLStreamError('stream argument needs to have a write() method')
		self._stream=val
	@property
	def serializer(self):
		try:
			if hasattr(self.dumper,'typ'):return self.dumper.serializer
			return self.dumper._serializer
		except AttributeError:return self
	@property
	def flow_level(self):return len(self.flow_context)
	def dispose(self):self.states=[];self.state=_B
	def emit(self,event):
		if dbg(DBG_EVENT):nprint(event)
		self.events.append(event)
		while not self.need_more_events():self.event=self.events.pop(0);self.state();self.event=_B
	def need_more_events(self):
		if not self.events:return _A
		event=self.events[0]
		if isinstance(event,DocumentStartEvent):return self.need_events(1)
		elif isinstance(event,SequenceStartEvent):return self.need_events(2)
		elif isinstance(event,MappingStartEvent):return self.need_events(3)
		else:return _C
	def need_events(self,count):
		level=0
		for event in self.events[1:]:
			if isinstance(event,(DocumentStartEvent,CollectionStartEvent)):level+=1
			elif isinstance(event,(DocumentEndEvent,CollectionEndEvent)):level-=1
			elif isinstance(event,StreamEndEvent):level=-1
			if level<0:return _C
		return len(self.events)<count+1
	def increase_indent(self,flow=_C,sequence=_B,indentless=_C):
		self.indents.append(self.indent,sequence)
		if self.indent is _B:
			if flow:self.indent=self.requested_indent
			else:self.indent=0
		elif not indentless:self.indent+=self.best_sequence_indent if self.indents.last_seq()else self.best_map_indent
	def expect_stream_start(self):
		A='encoding'
		if isinstance(self.event,StreamStartEvent):
			if PY2:
				if self.event.encoding and not getattr(self.stream,A,_B):self.encoding=self.event.encoding
			elif self.event.encoding and not hasattr(self.stream,A):self.encoding=self.event.encoding
			self.write_stream_start();self.state=self.expect_first_document_start
		else:raise EmitterError('expected StreamStartEvent, but got %s'%(self.event,))
	def expect_nothing(self):raise EmitterError('expected nothing, but got %s'%(self.event,))
	def expect_first_document_start(self):return self.expect_document_start(first=_A)
	def expect_document_start(self,first=_C):
		if isinstance(self.event,DocumentStartEvent):
			if(self.event.version or self.event.tags)and self.open_ended:self.write_indicator(_M,_A);self.write_indent()
			if self.event.version:version_text=self.prepare_version(self.event.version);self.write_version_directive(version_text)
			self.tag_prefixes=self.DEFAULT_TAG_PREFIXES.copy()
			if self.event.tags:
				handles=sorted(self.event.tags.keys())
				for handle in handles:prefix=self.event.tags[handle];self.tag_prefixes[prefix]=handle;handle_text=self.prepare_tag_handle(handle);prefix_text=self.prepare_tag_prefix(prefix);self.write_tag_directive(handle_text,prefix_text)
			implicit=first and not self.event.explicit and not self.canonical and not self.event.version and not self.event.tags and not self.check_empty_document()
			if not implicit:
				self.write_indent();self.write_indicator(_S,_A)
				if self.canonical:self.write_indent()
			self.state=self.expect_document_root
		elif isinstance(self.event,StreamEndEvent):
			if self.open_ended:self.write_indicator(_M,_A);self.write_indent()
			self.write_stream_end();self.state=self.expect_nothing
		else:raise EmitterError('expected DocumentStartEvent, but got %s'%(self.event,))
	def expect_document_end(self):
		if isinstance(self.event,DocumentEndEvent):
			self.write_indent()
			if self.event.explicit:self.write_indicator(_M,_A);self.write_indent()
			self.flush_stream();self.state=self.expect_document_start
		else:raise EmitterError('expected DocumentEndEvent, but got %s'%(self.event,))
	def expect_document_root(self):self.states.append(self.expect_document_end);self.expect_node(root=_A)
	def expect_node(self,root=_C,sequence=_C,mapping=_C,simple_key=_C):
		self.root_context=root;self.sequence_context=sequence;self.mapping_context=mapping;self.simple_key_context=simple_key
		if isinstance(self.event,AliasEvent):self.expect_alias()
		elif isinstance(self.event,(ScalarEvent,CollectionStartEvent)):
			if self.process_anchor('&')and isinstance(self.event,ScalarEvent)and self.sequence_context:self.sequence_context=_C
			if root and isinstance(self.event,ScalarEvent)and not self.scalar_after_indicator:self.write_indent()
			self.process_tag()
			if isinstance(self.event,ScalarEvent):self.expect_scalar()
			elif isinstance(self.event,SequenceStartEvent):
				i2,n2=self.indention,self.no_newline
				if self.event.comment:
					if self.event.flow_style is _C and self.event.comment:
						if self.write_post_comment(self.event):self.indention=_C;self.no_newline=_A
					if self.write_pre_comment(self.event):self.indention=i2;self.no_newline=not self.indention
				if self.flow_level or self.canonical or self.event.flow_style or self.check_empty_sequence():self.expect_flow_sequence()
				else:self.expect_block_sequence()
			elif isinstance(self.event,MappingStartEvent):
				if self.event.flow_style is _C and self.event.comment:self.write_post_comment(self.event)
				if self.event.comment and self.event.comment[1]:self.write_pre_comment(self.event)
				if self.flow_level or self.canonical or self.event.flow_style or self.check_empty_mapping():self.expect_flow_mapping(single=self.event.nr_items==1)
				else:self.expect_block_mapping()
		else:raise EmitterError('expected NodeEvent, but got %s'%(self.event,))
	def expect_alias(self):
		if self.event.anchor is _B:raise EmitterError('anchor is not specified for alias')
		self.process_anchor('*');self.state=self.states.pop()
	def expect_scalar(self):self.increase_indent(flow=_A);self.process_scalar();self.indent=self.indents.pop();self.state=self.states.pop()
	def expect_flow_sequence(self):ind=self.indents.seq_flow_align(self.best_sequence_indent,self.column);self.write_indicator(_D*ind+_K,_A,whitespace=_A);self.increase_indent(flow=_A,sequence=_A);self.flow_context.append(_K);self.state=self.expect_first_flow_sequence_item
	def expect_first_flow_sequence_item(self):
		if isinstance(self.event,SequenceEndEvent):
			self.indent=self.indents.pop();popped=self.flow_context.pop();assert popped==_K;self.write_indicator(']',_C)
			if self.event.comment and self.event.comment[0]:self.write_post_comment(self.event)
			elif self.flow_level==0:self.write_line_break()
			self.state=self.states.pop()
		else:
			if self.canonical or self.column>self.best_width:self.write_indent()
			self.states.append(self.expect_flow_sequence_item);self.expect_node(sequence=_A)
	def expect_flow_sequence_item(self):
		if isinstance(self.event,SequenceEndEvent):
			self.indent=self.indents.pop();popped=self.flow_context.pop();assert popped==_K
			if self.canonical:self.write_indicator(_N,_C);self.write_indent()
			self.write_indicator(']',_C)
			if self.event.comment and self.event.comment[0]:self.write_post_comment(self.event)
			else:self.no_newline=_C
			self.state=self.states.pop()
		else:
			self.write_indicator(_N,_C)
			if self.canonical or self.column>self.best_width:self.write_indent()
			self.states.append(self.expect_flow_sequence_item);self.expect_node(sequence=_A)
	def expect_flow_mapping(self,single=_C):
		ind=self.indents.seq_flow_align(self.best_sequence_indent,self.column);map_init='{'
		if single and self.flow_level and self.flow_context[-1]==_K and not self.canonical and not self.brace_single_entry_mapping_in_flow_sequence:map_init=''
		self.write_indicator(_D*ind+map_init,_A,whitespace=_A);self.flow_context.append(map_init);self.increase_indent(flow=_A,sequence=_C);self.state=self.expect_first_flow_mapping_key
	def expect_first_flow_mapping_key(self):
		if isinstance(self.event,MappingEndEvent):
			self.indent=self.indents.pop();popped=self.flow_context.pop();assert popped=='{';self.write_indicator('}',_C)
			if self.event.comment and self.event.comment[0]:self.write_post_comment(self.event)
			elif self.flow_level==0:self.write_line_break()
			self.state=self.states.pop()
		else:
			if self.canonical or self.column>self.best_width:self.write_indent()
			if not self.canonical and self.check_simple_key():self.states.append(self.expect_flow_mapping_simple_value);self.expect_node(mapping=_A,simple_key=_A)
			else:self.write_indicator(_I,_A);self.states.append(self.expect_flow_mapping_value);self.expect_node(mapping=_A)
	def expect_flow_mapping_key(self):
		if isinstance(self.event,MappingEndEvent):
			self.indent=self.indents.pop();popped=self.flow_context.pop();assert popped in['{','']
			if self.canonical:self.write_indicator(_N,_C);self.write_indent()
			if popped!='':self.write_indicator('}',_C)
			if self.event.comment and self.event.comment[0]:self.write_post_comment(self.event)
			else:self.no_newline=_C
			self.state=self.states.pop()
		else:
			self.write_indicator(_N,_C)
			if self.canonical or self.column>self.best_width:self.write_indent()
			if not self.canonical and self.check_simple_key():self.states.append(self.expect_flow_mapping_simple_value);self.expect_node(mapping=_A,simple_key=_A)
			else:self.write_indicator(_I,_A);self.states.append(self.expect_flow_mapping_value);self.expect_node(mapping=_A)
	def expect_flow_mapping_simple_value(self):self.write_indicator(self.prefixed_colon,_C);self.states.append(self.expect_flow_mapping_key);self.expect_node(mapping=_A)
	def expect_flow_mapping_value(self):
		if self.canonical or self.column>self.best_width:self.write_indent()
		self.write_indicator(self.prefixed_colon,_A);self.states.append(self.expect_flow_mapping_key);self.expect_node(mapping=_A)
	def expect_block_sequence(self):
		if self.mapping_context:indentless=not self.indention
		else:
			indentless=_C
			if not self.compact_seq_seq and self.column!=0:self.write_line_break()
		self.increase_indent(flow=_C,sequence=_A,indentless=indentless);self.state=self.expect_first_block_sequence_item
	def expect_first_block_sequence_item(self):return self.expect_block_sequence_item(first=_A)
	def expect_block_sequence_item(self,first=_C):
		if not first and isinstance(self.event,SequenceEndEvent):
			if self.event.comment and self.event.comment[1]:self.write_pre_comment(self.event)
			self.indent=self.indents.pop();self.state=self.states.pop();self.no_newline=_C
		else:
			if self.event.comment and self.event.comment[1]:self.write_pre_comment(self.event)
			nonl=self.no_newline if self.column==0 else _C;self.write_indent();ind=self.sequence_dash_offset;self.write_indicator(_D*ind+'-',_A,indention=_A)
			if nonl or self.sequence_dash_offset+2>self.best_sequence_indent:self.no_newline=_A
			self.states.append(self.expect_block_sequence_item);self.expect_node(sequence=_A)
	def expect_block_mapping(self):
		if not self.mapping_context and not(self.compact_seq_map or self.column==0):self.write_line_break()
		self.increase_indent(flow=_C,sequence=_C);self.state=self.expect_first_block_mapping_key
	def expect_first_block_mapping_key(self):return self.expect_block_mapping_key(first=_A)
	def expect_block_mapping_key(self,first=_C):
		if not first and isinstance(self.event,MappingEndEvent):
			if self.event.comment and self.event.comment[1]:self.write_pre_comment(self.event)
			self.indent=self.indents.pop();self.state=self.states.pop()
		else:
			if self.event.comment and self.event.comment[1]:self.write_pre_comment(self.event)
			self.write_indent()
			if self.check_simple_key():
				if not isinstance(self.event,(SequenceStartEvent,MappingStartEvent)):
					try:
						if self.event.style==_I:self.write_indicator(_I,_A,indention=_A)
					except AttributeError:pass
				self.states.append(self.expect_block_mapping_simple_value);self.expect_node(mapping=_A,simple_key=_A)
				if isinstance(self.event,AliasEvent):self.stream.write(_D)
			else:self.write_indicator(_I,_A,indention=_A);self.states.append(self.expect_block_mapping_value);self.expect_node(mapping=_A)
	def expect_block_mapping_simple_value(self):
		if getattr(self.event,'style',_B)!=_I:
			if self.indent==0 and self.top_level_colon_align is not _B:c=_D*(self.top_level_colon_align-self.column)+self.colon
			else:c=self.prefixed_colon
			self.write_indicator(c,_C)
		self.states.append(self.expect_block_mapping_key);self.expect_node(mapping=_A)
	def expect_block_mapping_value(self):self.write_indent();self.write_indicator(self.prefixed_colon,_A,indention=_A);self.states.append(self.expect_block_mapping_key);self.expect_node(mapping=_A)
	def check_empty_sequence(self):return isinstance(self.event,SequenceStartEvent)and bool(self.events)and isinstance(self.events[0],SequenceEndEvent)
	def check_empty_mapping(self):return isinstance(self.event,MappingStartEvent)and bool(self.events)and isinstance(self.events[0],MappingEndEvent)
	def check_empty_document(self):
		if not isinstance(self.event,DocumentStartEvent)or not self.events:return _C
		event=self.events[0];return isinstance(event,ScalarEvent)and event.anchor is _B and event.tag is _B and event.implicit and event.value==''
	def check_simple_key(self):
		length=0
		if isinstance(self.event,NodeEvent)and self.event.anchor is not _B:
			if self.prepared_anchor is _B:self.prepared_anchor=self.prepare_anchor(self.event.anchor)
			length+=len(self.prepared_anchor)
		if isinstance(self.event,(ScalarEvent,CollectionStartEvent))and self.event.tag is not _B:
			if self.prepared_tag is _B:self.prepared_tag=self.prepare_tag(self.event.tag)
			length+=len(self.prepared_tag)
		if isinstance(self.event,ScalarEvent):
			if self.analysis is _B:self.analysis=self.analyze_scalar(self.event.value)
			length+=len(self.analysis.scalar)
		return length<self.MAX_SIMPLE_KEY_LENGTH and(isinstance(self.event,AliasEvent)or isinstance(self.event,SequenceStartEvent)and self.event.flow_style is _A or isinstance(self.event,MappingStartEvent)and self.event.flow_style is _A or isinstance(self.event,ScalarEvent)and not(self.analysis.empty and self.style and self.style not in'\'"')and not self.analysis.multiline or self.check_empty_sequence()or self.check_empty_mapping())
	def process_anchor(self,indicator):
		if self.event.anchor is _B:self.prepared_anchor=_B;return _C
		if self.prepared_anchor is _B:self.prepared_anchor=self.prepare_anchor(self.event.anchor)
		if self.prepared_anchor:self.write_indicator(indicator+self.prepared_anchor,_A);self.no_newline=_C
		self.prepared_anchor=_B;return _A
	def process_tag(self):
		tag=self.event.tag
		if isinstance(self.event,ScalarEvent):
			if self.style is _B:self.style=self.choose_scalar_style()
			if(not self.canonical or tag is _B)and(self.style==''and self.event.implicit[0]or self.style!=''and self.event.implicit[1]):self.prepared_tag=_B;return
			if self.event.implicit[0]and tag is _B:tag=_G;self.prepared_tag=_B
		elif(not self.canonical or tag is _B)and self.event.implicit:self.prepared_tag=_B;return
		if tag is _B:raise EmitterError('tag is not specified')
		if self.prepared_tag is _B:self.prepared_tag=self.prepare_tag(tag)
		if self.prepared_tag:
			self.write_indicator(self.prepared_tag,_A)
			if self.sequence_context and not self.flow_level and isinstance(self.event,ScalarEvent):self.no_newline=_A
		self.prepared_tag=_B
	def choose_scalar_style(self):
		if self.analysis is _B:self.analysis=self.analyze_scalar(self.event.value)
		if self.event.style==_H or self.canonical:return _H
		if(not self.event.style or self.event.style==_I)and(self.event.implicit[0]or not self.event.implicit[2]):
			if not(self.simple_key_context and(self.analysis.empty or self.analysis.multiline))and(self.flow_level and self.analysis.allow_flow_plain or not self.flow_level and self.analysis.allow_block_plain):return''
		self.analysis.allow_block=_A
		if self.event.style and self.event.style in'|>':
			if not self.flow_level and not self.simple_key_context and self.analysis.allow_block:return self.event.style
		if not self.event.style and self.analysis.allow_double_quoted:
			if _J in self.event.value or _E in self.event.value:return _H
		if not self.event.style or self.event.style==_J:
			if self.analysis.allow_single_quoted and not(self.simple_key_context and self.analysis.multiline):return _J
		return _H
	def process_scalar(self):
		if self.analysis is _B:self.analysis=self.analyze_scalar(self.event.value)
		if self.style is _B:self.style=self.choose_scalar_style()
		split=not self.simple_key_context
		if self.sequence_context and not self.flow_level:self.write_indent()
		if self.style==_H:self.write_double_quoted(self.analysis.scalar,split)
		elif self.style==_J:self.write_single_quoted(self.analysis.scalar,split)
		elif self.style=='>':self.write_folded(self.analysis.scalar)
		elif self.style=='|':self.write_literal(self.analysis.scalar,self.event.comment)
		else:self.write_plain(self.analysis.scalar,split)
		self.analysis=_B;self.style=_B
		if self.event.comment:self.write_post_comment(self.event)
	def prepare_version(self,version):
		major,minor=version
		if major!=1:raise EmitterError('unsupported YAML version: %d.%d'%(major,minor))
		return'%d.%d'%(major,minor)
	def prepare_tag_handle(self,handle):
		if not handle:raise EmitterError('tag handle must not be empty')
		if handle[0]!=_G or handle[-1]!=_G:raise EmitterError("tag handle must start and end with '!': %r"%utf8(handle))
		for ch in handle[1:-1]:
			if not(_O<=ch<='9'or'A'<=ch<='Z'or _P<=ch<='z'or ch in'-_'):raise EmitterError('invalid character %r in the tag handle: %r'%(utf8(ch),utf8(handle)))
		return handle
	def prepare_tag_prefix(self,prefix):
		if not prefix:raise EmitterError('tag prefix must not be empty')
		chunks=[];start=end=0
		if prefix[0]==_G:end=1
		ch_set=_T
		if self.dumper:
			version=getattr(self.dumper,_U,(1,2))
			if version is _B or version>=(1,2):ch_set+='#'
		while end<len(prefix):
			ch=prefix[end]
			if _O<=ch<='9'or'A'<=ch<='Z'or _P<=ch<='z'or ch in ch_set:end+=1
			else:
				if start<end:chunks.append(prefix[start:end])
				start=end=end+1;data=utf8(ch)
				for ch in data:chunks.append(_V%ord(ch))
		if start<end:chunks.append(prefix[start:end])
		return ''.join(chunks)
	def prepare_tag(self,tag):
		if not tag:raise EmitterError('tag must not be empty')
		if tag==_G:return tag
		handle=_B;suffix=tag;prefixes=sorted(self.tag_prefixes.keys())
		for prefix in prefixes:
			if tag.startswith(prefix)and(prefix==_G or len(prefix)<len(tag)):handle=self.tag_prefixes[prefix];suffix=tag[len(prefix):]
		chunks=[];start=end=0;ch_set=_T
		if self.dumper:
			version=getattr(self.dumper,_U,(1,2))
			if version is _B or version>=(1,2):ch_set+='#'
		while end<len(suffix):
			ch=suffix[end]
			if _O<=ch<='9'or'A'<=ch<='Z'or _P<=ch<='z'or ch in ch_set or ch==_G and handle!=_G:end+=1
			else:
				if start<end:chunks.append(suffix[start:end])
				start=end=end+1;data=utf8(ch)
				for ch in data:chunks.append(_V%ord(ch))
		if start<end:chunks.append(suffix[start:end])
		suffix_text=''.join(chunks)
		if handle:return'%s%s'%(handle,suffix_text)
		else:return'!<%s>'%suffix_text
	def prepare_anchor(self,anchor):
		if not anchor:raise EmitterError('anchor must not be empty')
		for ch in anchor:
			if not check_anchorname_char(ch):raise EmitterError('invalid character %r in the anchor: %r'%(utf8(ch),utf8(anchor)))
		return anchor
	def analyze_scalar(self,scalar):
		A='\x00 \t\r\n\x85\u2028\u2029'
		if not scalar:return ScalarAnalysis(scalar=scalar,empty=_A,multiline=_C,allow_flow_plain=_C,allow_block_plain=_A,allow_single_quoted=_A,allow_double_quoted=_A,allow_block=_C)
		block_indicators=_C;flow_indicators=_C;line_breaks=_C;special_characters=_C;leading_space=_C;leading_break=_C;trailing_space=_C;trailing_break=_C;break_space=_C;space_break=_C
		if scalar.startswith(_S)or scalar.startswith(_M):block_indicators=_A;flow_indicators=_A
		preceeded_by_whitespace=_A;followed_by_whitespace=len(scalar)==1 or scalar[1]in A;previous_space=_C;previous_break=_C;index=0
		while index<len(scalar):
			ch=scalar[index]
			if index==0:
				if ch in'#,[]{}&*!|>\'"%@`':flow_indicators=_A;block_indicators=_A
				if ch in'?:':
					if self.serializer.use_version==(1,1):flow_indicators=_A
					elif len(scalar)==1:flow_indicators=_A
					if followed_by_whitespace:block_indicators=_A
				if ch=='-'and followed_by_whitespace:flow_indicators=_A;block_indicators=_A
			else:
				if ch in',[]{}':flow_indicators=_A
				if ch==_I and self.serializer.use_version==(1,1):flow_indicators=_A
				if ch==':':
					if followed_by_whitespace:flow_indicators=_A;block_indicators=_A
				if ch=='#'and preceeded_by_whitespace:flow_indicators=_A;block_indicators=_A
			if ch in _F:line_breaks=_A
			if not(ch==_E or _D<=ch<='~'):
				if(ch==_W or _Q<=ch<=_X or _Y<=ch<='�'or self.unicode_supplementary and'𐀀'<=ch<='\U0010ffff')and ch!=_Z:
					if not self.allow_unicode:special_characters=_A
				else:special_characters=_A
			if ch==_D:
				if index==0:leading_space=_A
				if index==len(scalar)-1:trailing_space=_A
				if previous_break:break_space=_A
				previous_space=_A;previous_break=_C
			elif ch in _F:
				if index==0:leading_break=_A
				if index==len(scalar)-1:trailing_break=_A
				if previous_space:space_break=_A
				previous_space=_C;previous_break=_A
			else:previous_space=_C;previous_break=_C
			index+=1;preceeded_by_whitespace=ch in A;followed_by_whitespace=index+1>=len(scalar)or scalar[index+1]in A
		allow_flow_plain=_A;allow_block_plain=_A;allow_single_quoted=_A;allow_double_quoted=_A;allow_block=_A
		if leading_space or leading_break or trailing_space or trailing_break:allow_flow_plain=allow_block_plain=_C
		if trailing_space:allow_block=_C
		if break_space:allow_flow_plain=allow_block_plain=allow_single_quoted=_C
		if special_characters:allow_flow_plain=allow_block_plain=allow_single_quoted=allow_block=_C
		elif space_break:
			allow_flow_plain=allow_block_plain=allow_single_quoted=_C
			if not self.allow_space_break:allow_block=_C
		if line_breaks:allow_flow_plain=allow_block_plain=_C
		if flow_indicators:allow_flow_plain=_C
		if block_indicators:allow_block_plain=_C
		return ScalarAnalysis(scalar=scalar,empty=_C,multiline=line_breaks,allow_flow_plain=allow_flow_plain,allow_block_plain=allow_block_plain,allow_single_quoted=allow_single_quoted,allow_double_quoted=allow_double_quoted,allow_block=allow_block)
	def flush_stream(self):
		if hasattr(self.stream,'flush'):self.stream.flush()
	def write_stream_start(self):
		if self.encoding and self.encoding.startswith('utf-16'):self.stream.write(_Z.encode(self.encoding))
	def write_stream_end(self):self.flush_stream()
	def write_indicator(self,indicator,need_whitespace,whitespace=_C,indention=_C):
		if self.whitespace or not need_whitespace:data=indicator
		else:data=_D+indicator
		self.whitespace=whitespace;self.indention=self.indention and indention;self.column+=len(data);self.open_ended=_C
		if bool(self.encoding):data=data.encode(self.encoding)
		self.stream.write(data)
	def write_indent(self):
		indent=self.indent or 0
		if not self.indention or self.column>indent or self.column==indent and not self.whitespace:
			if bool(self.no_newline):self.no_newline=_C
			else:self.write_line_break()
		if self.column<indent:
			self.whitespace=_A;data=_D*(indent-self.column);self.column=indent
			if self.encoding:data=data.encode(self.encoding)
			self.stream.write(data)
	def write_line_break(self,data=_B):
		if data is _B:data=self.best_line_break
		self.whitespace=_A;self.indention=_A;self.line+=1;self.column=0
		if bool(self.encoding):data=data.encode(self.encoding)
		self.stream.write(data)
	def write_version_directive(self,version_text):
		data='%%YAML %s'%version_text
		if self.encoding:data=data.encode(self.encoding)
		self.stream.write(data);self.write_line_break()
	def write_tag_directive(self,handle_text,prefix_text):
		data='%%TAG %s %s'%(handle_text,prefix_text)
		if self.encoding:data=data.encode(self.encoding)
		self.stream.write(data);self.write_line_break()
	def write_single_quoted(self,text,split=_A):
		if self.root_context:
			if self.requested_indent is not _B:
				self.write_line_break()
				if self.requested_indent!=0:self.write_indent()
		self.write_indicator(_J,_A);spaces=_C;breaks=_C;start=end=0
		while end<=len(text):
			ch=_B
			if end<len(text):ch=text[end]
			if spaces:
				if ch is _B or ch!=_D:
					if start+1==end and self.column>self.best_width and split and start!=0 and end!=len(text):self.write_indent()
					else:
						data=text[start:end];self.column+=len(data)
						if bool(self.encoding):data=data.encode(self.encoding)
						self.stream.write(data)
					start=end
			elif breaks:
				if ch is _B or ch not in _F:
					if text[start]==_E:self.write_line_break()
					for br in text[start:end]:
						if br==_E:self.write_line_break()
						else:self.write_line_break(br)
					self.write_indent();start=end
			elif ch is _B or ch in _R or ch==_J:
				if start<end:
					data=text[start:end];self.column+=len(data)
					if bool(self.encoding):data=data.encode(self.encoding)
					self.stream.write(data);start=end
			if ch==_J:
				data="''";self.column+=2
				if bool(self.encoding):data=data.encode(self.encoding)
				self.stream.write(data);start=end+1
			if ch is not _B:spaces=ch==_D;breaks=ch in _F
			end+=1
		self.write_indicator(_J,_C)
	ESCAPE_REPLACEMENTS={'\x00':_O,_a:_P,'\x08':'b','\t':'t',_E:'n','\x0b':'v','\x0c':'f','\r':'r','\x1b':'e',_H:_H,_L:_L,_W:'N',_Q:'_','\u2028':'L','\u2029':'P'}
	def write_double_quoted(self,text,split=_A):
		if self.root_context:
			if self.requested_indent is not _B:
				self.write_line_break()
				if self.requested_indent!=0:self.write_indent()
		self.write_indicator(_H,_A);start=end=0
		while end<=len(text):
			ch=_B
			if end<len(text):ch=text[end]
			if ch is _B or ch in'"\\\x85\u2028\u2029\ufeff'or not(_D<=ch<='~'or self.allow_unicode and(_Q<=ch<=_X or _Y<=ch<='�')):
				if start<end:
					data=text[start:end];self.column+=len(data)
					if bool(self.encoding):data=data.encode(self.encoding)
					self.stream.write(data);start=end
				if ch is not _B:
					if ch in self.ESCAPE_REPLACEMENTS:data=_L+self.ESCAPE_REPLACEMENTS[ch]
					elif ch<='ÿ':data='\\x%02X'%ord(ch)
					elif ch<='\uffff':data='\\u%04X'%ord(ch)
					else:data='\\U%08X'%ord(ch)
					self.column+=len(data)
					if bool(self.encoding):data=data.encode(self.encoding)
					self.stream.write(data);start=end+1
			if 0<end<len(text)-1 and(ch==_D or start>=end)and self.column+(end-start)>self.best_width and split:
				data=text[start:end]+_L
				if start<end:start=end
				self.column+=len(data)
				if bool(self.encoding):data=data.encode(self.encoding)
				self.stream.write(data);self.write_indent();self.whitespace=_C;self.indention=_C
				if text[start]==_D:
					data=_L;self.column+=len(data)
					if bool(self.encoding):data=data.encode(self.encoding)
					self.stream.write(data)
			end+=1
		self.write_indicator(_H,_C)
	def determine_block_hints(self,text):
		indent=0;indicator='';hints=''
		if text:
			if text[0]in _R:indent=self.best_sequence_indent;hints+=text_type(indent)
			elif self.root_context:
				for end in ['\n---','\n...']:
					pos=0
					while _A:
						pos=text.find(end,pos)
						if pos==-1:break
						try:
							if text[pos+4]in' \r\n':break
						except IndexError:pass
						pos+=1
					if pos>-1:break
				if pos>0:indent=self.best_sequence_indent
			if text[-1]not in _F:indicator='-'
			elif len(text)==1 or text[-2]in _F:indicator='+'
		hints+=indicator;return hints,indent,indicator
	def write_folded(self,text):
		hints,_indent,_indicator=self.determine_block_hints(text);self.write_indicator('>'+hints,_A)
		if _indicator=='+':self.open_ended=_A
		self.write_line_break();leading_space=_A;spaces=_C;breaks=_A;start=end=0
		while end<=len(text):
			ch=_B
			if end<len(text):ch=text[end]
			if breaks:
				if ch is _B or ch not in'\n\x85\u2028\u2029\x07':
					if not leading_space and ch is not _B and ch!=_D and text[start]==_E:self.write_line_break()
					leading_space=ch==_D
					for br in text[start:end]:
						if br==_E:self.write_line_break()
						else:self.write_line_break(br)
					if ch is not _B:self.write_indent()
					start=end
			elif spaces:
				if ch!=_D:
					if start+1==end and self.column>self.best_width:self.write_indent()
					else:
						data=text[start:end];self.column+=len(data)
						if bool(self.encoding):data=data.encode(self.encoding)
						self.stream.write(data)
					start=end
			elif ch is _B or ch in' \n\x85\u2028\u2029\x07':
				data=text[start:end];self.column+=len(data)
				if bool(self.encoding):data=data.encode(self.encoding)
				self.stream.write(data)
				if ch==_a:
					if end<len(text)-1 and not text[end+2].isspace():self.write_line_break();self.write_indent();end+=2
					else:raise EmitterError('unexcpected fold indicator \\a before space')
				if ch is _B:self.write_line_break()
				start=end
			if ch is not _B:breaks=ch in _F;spaces=ch==_D
			end+=1
	def write_literal(self,text,comment=_B):
		hints,_indent,_indicator=self.determine_block_hints(text);self.write_indicator('|'+hints,_A)
		try:
			comment=comment[1][0]
			if comment:self.stream.write(comment)
		except (TypeError,IndexError):pass
		if _indicator=='+':self.open_ended=_A
		self.write_line_break();breaks=_A;start=end=0
		while end<=len(text):
			ch=_B
			if end<len(text):ch=text[end]
			if breaks:
				if ch is _B or ch not in _F:
					for br in text[start:end]:
						if br==_E:self.write_line_break()
						else:self.write_line_break(br)
					if ch is not _B:
						if self.root_context:idnx=self.indent if self.indent is not _B else 0;self.stream.write(_D*(_indent+idnx))
						else:self.write_indent()
					start=end
			elif ch is _B or ch in _F:
				data=text[start:end]
				if bool(self.encoding):data=data.encode(self.encoding)
				self.stream.write(data)
				if ch is _B:self.write_line_break()
				start=end
			if ch is not _B:breaks=ch in _F
			end+=1
	def write_plain(self,text,split=_A):
		if self.root_context:
			if self.requested_indent is not _B:
				self.write_line_break()
				if self.requested_indent!=0:self.write_indent()
			else:self.open_ended=_A
		if not text:return
		if not self.whitespace:
			data=_D;self.column+=len(data)
			if self.encoding:data=data.encode(self.encoding)
			self.stream.write(data)
		self.whitespace=_C;self.indention=_C;spaces=_C;breaks=_C;start=end=0
		while end<=len(text):
			ch=_B
			if end<len(text):ch=text[end]
			if spaces:
				if ch!=_D:
					if start+1==end and self.column>self.best_width and split:self.write_indent();self.whitespace=_C;self.indention=_C
					else:
						data=text[start:end];self.column+=len(data)
						if self.encoding:data=data.encode(self.encoding)
						self.stream.write(data)
					start=end
			elif breaks:
				if ch not in _F:
					if text[start]==_E:self.write_line_break()
					for br in text[start:end]:
						if br==_E:self.write_line_break()
						else:self.write_line_break(br)
					self.write_indent();self.whitespace=_C;self.indention=_C;start=end
			elif ch is _B or ch in _R:
				data=text[start:end];self.column+=len(data)
				if self.encoding:data=data.encode(self.encoding)
				try:self.stream.write(data)
				except:sys.stdout.write(repr(data)+_E);raise
				start=end
			if ch is not _B:spaces=ch==_D;breaks=ch in _F
			end+=1
	def write_comment(self,comment,pre=_C):
		value=comment.value
		if not pre and value[-1]==_E:value=value[:-1]
		try:
			col=comment.start_mark.column
			if comment.value and comment.value.startswith(_E):col=self.column
			elif col<self.column+1:ValueError
		except ValueError:col=self.column+1
		try:
			nr_spaces=col-self.column
			if self.column and value.strip()and nr_spaces<1 and value[0]!=_E:nr_spaces=1
			value=_D*nr_spaces+value
			try:
				if bool(self.encoding):value=value.encode(self.encoding)
			except UnicodeDecodeError:pass
			self.stream.write(value)
		except TypeError:raise
		if not pre:self.write_line_break()
	def write_pre_comment(self,event):
		comments=event.comment[1]
		if comments is _B:return _C
		try:
			start_events=MappingStartEvent,SequenceStartEvent
			for comment in comments:
				if isinstance(event,start_events)and getattr(comment,'pre_done',_B):continue
				if self.column!=0:self.write_line_break()
				self.write_comment(comment,pre=_A)
				if isinstance(event,start_events):comment.pre_done=_A
		except TypeError:sys.stdout.write('eventtt {} {}'.format(type(event),event));raise
		return _A
	def write_post_comment(self,event):
		if self.event.comment[0]is _B:return _C
		comment=event.comment[0];self.write_comment(comment);return _A


_F='always'
_E='key'
_D='%s=%s'
_C='QUOTE'
_B='FILE'
_A=True
import os,sys
from subprocess import Popen
try:from dynaconf.vendor import click
except ImportError:sys.stderr.write('It seems python-dotenv is not installed with cli option. \nRun pip install "python-dotenv[cli]" to fix this.');sys.exit(1)
from .compat import IS_TYPE_CHECKING,to_env
from .main import dotenv_values,get_key,set_key,unset_key
from .version import __version__
if IS_TYPE_CHECKING:from typing import Any,List,Dict
@click.group()
@click.option('-f','--file',default=os.path.join(os.getcwd(),'.env'),type=click.Path(exists=_A),help='Location of the .env file, defaults to .env file in current working directory.')
@click.option('-q','--quote',default=_F,type=click.Choice([_F,'never','auto']),help='Whether to quote or not the variable values. Default mode is always. This does not affect parsing.')
@click.version_option(version=__version__)
@click.pass_context
def cli(ctx,file,quote):A=ctx;A.obj={};A.obj[_B]=file;A.obj[_C]=quote
@cli.command()
@click.pass_context
def list(ctx):
	A=ctx.obj[_B];B=dotenv_values(A)
	for (C,D) in B.items():click.echo(_D%(C,D))
@cli.command()
@click.pass_context
@click.argument(_E,required=_A)
@click.argument('value',required=_A)
def set(ctx,key,value):
	B=value;A=key;C=ctx.obj[_B];D=ctx.obj[_C];E,A,B=set_key(C,A,B,D)
	if E:click.echo(_D%(A,B))
	else:exit(1)
@cli.command()
@click.pass_context
@click.argument(_E,required=_A)
def get(ctx,key):
	B=ctx.obj[_B];A=get_key(B,key)
	if A:click.echo(_D%(key,A))
	else:exit(1)
@cli.command()
@click.pass_context
@click.argument(_E,required=_A)
def unset(ctx,key):
	A=key;B=ctx.obj[_B];C=ctx.obj[_C];D,A=unset_key(B,A,C)
	if D:click.echo('Successfully removed %s'%A)
	else:exit(1)
@cli.command(context_settings={'ignore_unknown_options':_A})
@click.pass_context
@click.argument('commandline',nargs=-1,type=click.UNPROCESSED)
def run(ctx,commandline):
	A=commandline;B=ctx.obj[_B];C={to_env(C):to_env(A)for(C,A)in dotenv_values(B).items()if A is not None}
	if not A:click.echo('No command given.');exit(1)
	D=run_command(A,C);exit(D)
def run_command(command,env):A=os.environ.copy();A.update(env);B=Popen(command,universal_newlines=_A,bufsize=0,shell=False,env=A);C,C=B.communicate();return B.returncode
if __name__=='__main__':cli()


import glob
import importlib
import inspect
import os
import warnings
from contextlib import contextmanager
from contextlib import suppress

from dynaconf import default_settings
from dynaconf.loaders import default_loader
from dynaconf.loaders import enable_external_loaders
from dynaconf.loaders import env_loader
from dynaconf.loaders import py_loader
from dynaconf.loaders import settings_loader
from dynaconf.loaders import yaml_loader
from dynaconf.utils import BANNER
from dynaconf.utils import compat_kwargs
from dynaconf.utils import ensure_a_list
from dynaconf.utils import missing
from dynaconf.utils import object_merge
from dynaconf.utils import recursively_evaluate_lazy_format
from dynaconf.utils import RENAMED_VARS
from dynaconf.utils import upperfy
from dynaconf.utils.boxing import DynaBox
from dynaconf.utils.files import find_file
from dynaconf.utils.functional import empty
from dynaconf.utils.functional import LazyObject
from dynaconf.utils.parse_conf import converters
from dynaconf.utils.parse_conf import get_converter
from dynaconf.utils.parse_conf import parse_conf_data
from dynaconf.utils.parse_conf import true_values
from dynaconf.validator import ValidatorList


class LazySettings(LazyObject):
    """When you do::

        >>> from dynaconf import Dynaconf
        >>> settings = Dynaconf(*options)

    a LazySettings is initialized with only default_settings and all the
    parameters passed to *options.

    Then when you first access a value, this will be set up and loaders will
    be executed looking for defined config files or the file defined in
    SETTINGS_FILE_FOR_DYNACONF environment variable::

        $ export SETTINGS_FILE_FOR_DYNACONF=path

        >>> settings.FOO

    Or when you call::

        >>> settings.configure(settings_module='/tmp/settings.py')

    You can define in your settings module a list of loaders to get values
    from different stores. By default it will try environment variables
    starting with ENVVAR_PREFIX_FOR_DYNACONF (by defaulf `DYNACONF_`)

    You can also customize specific parameters.

    Exmaple: `proj/config.py`::

        >>> from dynaconf import Dynaconf
        >>> settings = Dynaconf(
        ...    env='production',
        ...    loaders=[
        ...        'dynaconf.loaders.env_loader',
        ...        'dynaconf.loaders.redis_loader'
        ...    ]
        ... )

    save common values in a settings file::

        $ echo "SERVER_IP = '10.10.10.10'" > proj/settings.py

    or use `.toml|.yaml|.ini|.json`

    save sensitive values in .secrets.{py|toml|yaml|ini|json}
    or export as DYNACONF global environment variable::

        $ export DYNACONF_SERVER_PASSWD='super_secret'

        >>> # from proj.config import settings
        >>> print settings.SERVER_IP
        >>> print settings.SERVER_PASSWD

    and now it reads all variables starting with `DYNACONF_` from envvars
    and all values in a hash called DYNACONF_PROJ in redis
    """

    def __init__(self, **kwargs):
        """
        handle initialization for the customization cases

        :param kwargs: values that overrides default_settings
        """

        self._warn_dynaconf_global_settings = kwargs.pop(
            "warn_dynaconf_global_settings", None
        )  # in 3.0.0 global settings is deprecated

        self.__resolve_config_aliases(kwargs)
        compat_kwargs(kwargs)
        self._kwargs = kwargs
        super(LazySettings, self).__init__()

    def __resolve_config_aliases(self, kwargs):
        """takes aliases for _FOR_DYNACONF configurations

        e.g: ROOT_PATH='/' is transformed into `ROOT_PATH_FOR_DYNACONF`
        """

        mispells = {
            "settings_files": "settings_file",
            "SETTINGS_FILES": "SETTINGS_FILE",
        }
        for mispell, correct in mispells.items():
            if mispell in kwargs:
                kwargs[correct] = kwargs.pop(mispell)

        for_dynaconf_keys = {
            key
            for key in dir(default_settings)
            if key.endswith("_FOR_DYNACONF")
        }
        aliases = {
            key.upper()
            for key in kwargs
            if f"{key.upper()}_FOR_DYNACONF" in for_dynaconf_keys
        }
        for alias in aliases:
            value = kwargs.pop(alias, empty)
            if value is empty:
                value = kwargs.pop(alias.lower())
            kwargs[f"{alias}_FOR_DYNACONF"] = value

    def __getattr__(self, name):
        """Allow getting keys from self.store using dot notation"""
        if self._wrapped is empty:
            self._setup()
        if name in self._wrapped._deleted:  # noqa
            raise AttributeError(
                f"Attribute {name} was deleted, " "or belongs to different env"
            )

        if name not in RESERVED_ATTRS:
            lowercase_mode = self._kwargs.get(
                "LOWERCASE_READ_FOR_DYNACONF",
                default_settings.LOWERCASE_READ_FOR_DYNACONF,
            )
            if lowercase_mode is True:
                name = name.upper()

        if (
            name.isupper()
            and (
                self._wrapped._fresh
                or name in self._wrapped.FRESH_VARS_FOR_DYNACONF
            )
            and name not in dir(default_settings)
        ):
            return self._wrapped.get_fresh(name)
        value = getattr(self._wrapped, name)
        if name not in RESERVED_ATTRS:
            return recursively_evaluate_lazy_format(value, self)
        return value

    def __call__(self, *args, **kwargs):
        """Allow direct call of settings('val')
        in place of settings.get('val')
        """
        return self.get(*args, **kwargs)

    def _setup(self):
        """Initial setup, run once."""

        if self._warn_dynaconf_global_settings:
            warnings.warn(
                "Usage of `from dynaconf import settings` is now "
                "DEPRECATED in 3.0.0+. You are encouraged to change it to "
                "your own instance e.g: `settings = Dynaconf(*options)`",
                DeprecationWarning,
            )

        default_settings.reload(self._kwargs.get("load_dotenv"))
        environment_variable = self._kwargs.get(
            "ENVVAR_FOR_DYNACONF", default_settings.ENVVAR_FOR_DYNACONF
        )
        settings_module = os.environ.get(environment_variable)
        self._wrapped = Settings(
            settings_module=settings_module, **self._kwargs
        )

    def configure(self, settings_module=None, **kwargs):
        """
        Allows user to reconfigure settings object passing a new settings
        module or separated kwargs

        :param settings_module: defines the setttings file
        :param kwargs:  override default settings
        """
        default_settings.reload(self._kwargs.get("load_dotenv"))
        environment_var = self._kwargs.get(
            "ENVVAR_FOR_DYNACONF", default_settings.ENVVAR_FOR_DYNACONF
        )
        settings_module = settings_module or os.environ.get(environment_var)
        compat_kwargs(kwargs)
        kwargs.update(self._kwargs)
        self._wrapped = Settings(settings_module=settings_module, **kwargs)

    @property
    def configured(self):
        """If wrapped is configured"""
        return self._wrapped is not empty


class Settings(object):
    """
    Common logic for settings whether set by a module or by the user.
    """

    dynaconf_banner = BANNER

    def __init__(self, settings_module=None, **kwargs):  # pragma: no cover
        """Execute loaders and custom initialization

        :param settings_module: defines the setttings file
        :param kwargs:  override default settings
        """
        self._fresh = False
        self._loaded_envs = []
        self._loaded_files = []
        self._deleted = set()
        self._store = DynaBox(box_settings=self)
        self._env_cache = {}
        self._loaded_by_loaders = {}
        self._loaders = []
        self._defaults = DynaBox(box_settings=self)
        self.environ = os.environ
        self.SETTINGS_MODULE = None
        self._not_installed_warnings = []
        self._memoized = None

        self.validators = ValidatorList(
            self, validators=kwargs.pop("validators", None)
        )

        compat_kwargs(kwargs)
        if settings_module:
            self.set("SETTINGS_FILE_FOR_DYNACONF", settings_module)
        for key, value in kwargs.items():
            self.set(key, value)
        # execute loaders only after setting defaults got from kwargs
        self._defaults = kwargs
        self.execute_loaders()

        self.validators.validate()

    def __call__(self, *args, **kwargs):
        """Allow direct call of `settings('val')`
        in place of `settings.get('val')`
        """
        return self.get(*args, **kwargs)

    def __setattr__(self, name, value):
        """Allow `settings.FOO = 'value'` while keeping internal attrs."""

        if name in RESERVED_ATTRS:
            super(Settings, self).__setattr__(name, value)
        else:
            self.set(name, value)

    def __delattr__(self, name):
        """stores reference in `_deleted` for proper error management"""
        self._deleted.add(name)
        if hasattr(self, name):
            super(Settings, self).__delattr__(name)

    def __contains__(self, item):
        "Respond to `item in settings`"
        return item.upper() in self.store or item.lower() in self.store

    def __getitem__(self, item):
        """Allow getting variables as dict keys `settings['KEY']`"""
        value = self.get(item, default=empty)
        if value is empty:
            raise KeyError(f"{item} does not exist")
        return value

    def __setitem__(self, key, value):
        """Allow `settings['KEY'] = 'value'`"""
        self.set(key, value)

    @property
    def store(self):
        """Gets internal storage"""
        return self._store

    def __dir__(self):
        """Enable auto-complete for code editors"""
        return (
            RESERVED_ATTRS
            + [k.lower() for k in self.keys()]
            + list(self.keys())
        )

    def __iter__(self):
        """Redirects to store object"""
        yield from self._store

    def items(self):
        """Redirects to store object"""
        return self._store.items()

    def keys(self):
        """Redirects to store object"""
        return self.store.keys()

    def values(self):
        """Redirects to store object"""
        return self.store.values()

    def setdefault(self, item, default):
        """Returns value if exists or set it as the given default"""
        value = self.get(item, empty)
        if value is empty and default is not empty:
            self.set(
                item,
                default,
                loader_identifier="setdefault",
                tomlfy=True,
                dotted_lookup=True,
            )
            return default
        return value

    def as_dict(self, env=None, internal=False):
        """Returns a dictionary with set key and values.

        :param env: Str env name, default self.current_env `DEVELOPMENT`
        :param internal: bool - should include dynaconf internal vars?
        """
        ctx_mgr = suppress() if env is None else self.using_env(env)
        with ctx_mgr:
            data = self.store.to_dict().copy()
            # if not internal remove internal settings
            if not internal:
                for name in dir(default_settings):
                    data.pop(name, None)
            return data

    to_dict = as_dict  # backwards compatibility

    def _dotted_get(self, dotted_key, default=None, parent=None, **kwargs):
        """
        Perform dotted key lookups and keep track of where we are.
        :param key: The name of the setting value, will always be upper case
        :param default: In case of not found it will be returned
        :param parent: Is there a pre-loaded parent in a nested data?
        """
        split_key = dotted_key.split(".")
        name, keys = split_key[0], split_key[1:]
        result = self.get(name, default=default, parent=parent, **kwargs)

        # If we've reached the end, or parent key not found, then return result
        if not keys or result == default:
            return result

        # If we've still got key elements to traverse, let's do that.
        return self._dotted_get(
            ".".join(keys), default=default, parent=result, **kwargs
        )

    def get(
        self,
        key,
        default=None,
        cast=None,
        fresh=False,
        dotted_lookup=True,
        parent=None,
    ):
        """
        Get a value from settings store, this is the prefered way to access::

            >>> from dynaconf import settings
            >>> settings.get('KEY')

        :param key: The name of the setting value, will always be upper case
        :param default: In case of not found it will be returned
        :param cast: Should cast in to @int, @float, @bool or @json ?
        :param fresh: Should reload from loaders store before access?
        :param dotted_lookup: Should perform dotted-path lookup?
        :param parent: Is there a pre-loaded parent in a nested data?
        :return: The value if found, default or None
        """

        if "." in key and dotted_lookup:
            return self._dotted_get(
                dotted_key=key,
                default=default,
                cast=cast,
                fresh=fresh,
                parent=parent,
            )

        key = upperfy(key)
        if key in self._deleted:
            return default

        if (
            fresh
            or self._fresh
            or key in getattr(self, "FRESH_VARS_FOR_DYNACONF", ())
        ) and key not in dir(default_settings):
            self.unset(key)
            self.execute_loaders(key=key)

        data = (parent or self.store).get(key, default)
        if cast:
            data = get_converter(cast, data, box_settings=self)
        return data

    def exists(self, key, fresh=False):
        """Check if key exists

        :param key: the name of setting variable
        :param fresh: if key should be taken from source direclty
        :return: Boolean
        """
        key = upperfy(key)
        if key in self._deleted:
            return False
        return self.get(key, fresh=fresh, default=missing) is not missing

    def get_fresh(self, key, default=None, cast=None):
        """This is a shortcut to `get(key, fresh=True)`. always reload from
        loaders store before getting the var.

        :param key: The name of the setting value, will always be upper case
        :param default: In case of not found it will be returned
        :param cast: Should cast in to @int, @float, @bool or @json ?
        :return: The value if found, default or None
        """
        return self.get(key, default=default, cast=cast, fresh=True)

    def get_environ(self, key, default=None, cast=None):
        """Get value from environment variable using os.environ.get

        :param key: The name of the setting value, will always be upper case
        :param default: In case of not found it will be returned
        :param cast: Should cast in to @int, @float, @bool or @json ?
         or cast must be true to use cast inference
        :return: The value if found, default or None
        """
        key = upperfy(key)
        data = self.environ.get(key, default)
        if data:
            if cast in converters:
                data = get_converter(cast, data, box_settings=self)
            if cast is True:
                data = parse_conf_data(data, tomlfy=True, box_settings=self)
        return data

    def exists_in_environ(self, key):
        """Return True if env variable is exported"""
        return upperfy(key) in self.environ

    def as_bool(self, key):
        """Partial method for get with bool cast"""
        return self.get(key, cast="@bool")

    def as_int(self, key):
        """Partial method for get with int cast"""
        return self.get(key, cast="@int")

    def as_float(self, key):
        """Partial method for get with float cast"""
        return self.get(key, cast="@float")

    def as_json(self, key):
        """Partial method for get with json cast"""
        return self.get(key, cast="@json")

    @property
    def loaded_envs(self):
        """Get or create internal loaded envs list"""
        if not self._loaded_envs:
            self._loaded_envs = []
        return self._loaded_envs

    # compat
    loaded_namespaces = loaded_envs

    @loaded_envs.setter
    def loaded_envs(self, value):
        """Setter for env list"""
        self._loaded_envs = value

    @property
    def loaded_by_loaders(self):
        """Gets the internal mapping of LOADER -> values"""
        return self._loaded_by_loaders

    def from_env(self, env="", keep=False, **kwargs):
        """Return a new isolated settings object pointing to specified env.

        Example of settings.toml::

            [development]
            message = 'This is in dev'
            [other]
            message = 'this is in other env'

        Program::

            >>> from dynaconf import settings
            >>> print(settings.MESSAGE)
            'This is in dev'
            >>> print(settings.from_env('other').MESSAGE)
            'This is in other env'
            # The existing settings object remains the same.
            >>> print(settings.MESSAGE)
            'This is in dev'

        Arguments:
            env {str} -- Env to load (development, production, custom)

        Keyword Arguments:
            keep {bool} -- Keep pre-existing values (default: {False})
            kwargs {dict} -- Passed directly to new instance.
        """
        cache_key = f"{env}_{keep}_{kwargs}"
        if cache_key in self._env_cache:
            return self._env_cache[cache_key]

        new_data = {
            key: self.get(key)
            for key in dir(default_settings)
            if key.isupper() and key not in RENAMED_VARS
        }

        # This is here for backwards compatibility
        # To be removed on 4.x.x
        default_settings_paths = self.get("default_settings_paths")
        if default_settings_paths:  # pragma: no cover
            new_data["default_settings_paths"] = default_settings_paths

        if keep:
            # keep existing values from current env
            new_data.update(
                {
                    key: value
                    for key, value in self.store.to_dict().copy().items()
                    if key.isupper() and key not in RENAMED_VARS
                }
            )

        new_data.update(kwargs)
        new_data["FORCE_ENV_FOR_DYNACONF"] = env
        new_settings = LazySettings(**new_data)
        self._env_cache[cache_key] = new_settings
        return new_settings

    @contextmanager
    def using_env(self, env, clean=True, silent=True, filename=None):
        """
        This context manager allows the contextual use of a different env
        Example of settings.toml::

            [development]
            message = 'This is in dev'
            [other]
            message = 'this is in other env'

        Program::

            >>> from dynaconf import settings
            >>> print settings.MESSAGE
            'This is in dev'
            >>> with settings.using_env('OTHER'):
            ...    print settings.MESSAGE
            'this is in other env'

        :param env: Upper case name of env without any _
        :param clean: If preloaded vars should be cleaned
        :param silent: Silence errors
        :param filename: Custom filename to load (optional)
        :return: context
        """
        try:
            self.setenv(env, clean=clean, silent=silent, filename=filename)
            yield
        finally:
            if env.lower() != self.ENV_FOR_DYNACONF.lower():
                del self.loaded_envs[-1]
            self.setenv(self.current_env, clean=clean, filename=filename)

    # compat
    using_namespace = using_env

    @contextmanager
    def fresh(self):
        """
        this context manager force the load of a key direct from the store::

            $ export DYNACONF_VALUE='Original'
            >>> from dynaconf import settings
            >>> print settings.VALUE
            'Original'
            $ export DYNACONF_VALUE='Changed Value'
            >>> print settings.VALUE  # will not be reloaded from env vars
            'Original
            >>> with settings.fresh(): # inside this context all is reloaded
            ...    print settings.VALUE
            'Changed Value'

        an alternative is using `settings.get_fresh(key)`

        :return: context
        """

        self._fresh = True
        yield
        self._fresh = False

    @property
    def current_env(self):
        """Return the current active env"""

        if self.ENVIRONMENTS_FOR_DYNACONF is False:
            return "main"

        if self.FORCE_ENV_FOR_DYNACONF is not None:
            return self.FORCE_ENV_FOR_DYNACONF

        try:
            return self.loaded_envs[-1]
        except IndexError:
            return self.ENV_FOR_DYNACONF

    # compat
    current_namespace = current_env

    @property
    def settings_module(self):
        """Gets SETTINGS_MODULE variable"""
        settings_module = parse_conf_data(
            os.environ.get(
                self.ENVVAR_FOR_DYNACONF, self.SETTINGS_FILE_FOR_DYNACONF
            ),
            tomlfy=True,
            box_settings=self,
        )
        if settings_module != getattr(self, "SETTINGS_MODULE", None):
            self.set("SETTINGS_MODULE", settings_module)

        # This is for backewards compatibility, to be removed on 4.x.x
        if not self.SETTINGS_MODULE and self.get("default_settings_paths"):
            self.SETTINGS_MODULE = self.get("default_settings_paths")

        return self.SETTINGS_MODULE

    # Backwards compatibility see #169
    settings_file = settings_module

    def setenv(self, env=None, clean=True, silent=True, filename=None):
        """Used to interactively change the env
        Example of settings.toml::

            [development]
            message = 'This is in dev'
            [other]
            message = 'this is in other env'

        Program::

            >>> from dynaconf import settings
            >>> print settings.MESSAGE
            'This is in dev'
            >>> with settings.using_env('OTHER'):
            ...    print settings.MESSAGE
            'this is in other env'

        :param env: Upper case name of env without any _
        :param clean: If preloaded vars should be cleaned
        :param silent: Silence errors
        :param filename: Custom filename to load (optional)
        :return: context
        """
        env = env or self.ENV_FOR_DYNACONF

        if not isinstance(env, str):
            raise AttributeError("env should be a string")

        env = env.upper()

        if env != self.ENV_FOR_DYNACONF:
            self.loaded_envs.append(env)
        else:
            self.loaded_envs = []

        if clean:
            self.clean(env=env)
        self.execute_loaders(env=env, silent=silent, filename=filename)

    # compat
    namespace = setenv

    def clean(self, *args, **kwargs):
        """Clean all loaded values to reload when switching envs"""
        for key in list(self.store.keys()):
            self.unset(key)

    def unset(self, key, force=False):
        """Unset on all references

        :param key: The key to be unset
        :param force: Bypass default checks and force unset
        """
        key = upperfy(key.strip())
        if (
            key not in dir(default_settings)
            and key not in self._defaults
            or force
        ):
            delattr(self, key)
            self.store.pop(key, None)

    def unset_all(self, keys, force=False):  # pragma: no cover
        """Unset based on a list of keys

        :param keys: a list of keys
        :param force: Bypass default checks and force unset
        """
        for key in keys:
            self.unset(key, force=force)

    def _dotted_set(self, dotted_key, value, tomlfy=False, **kwargs):
        """Sets dotted keys as nested dictionaries.

        Dotted set will always reassign the value, to merge use `@merge` token

        Arguments:
            dotted_key {str} -- A traversal name e.g: foo.bar.zaz
            value {Any} -- The value to set to the nested value.

        Keyword Arguments:
            tomlfy {bool} -- Perform toml parsing (default: {False})
        """

        split_keys = dotted_key.split(".")
        existing_data = self.get(split_keys[0], {})
        new_data = tree = DynaBox(box_settings=self)

        for k in split_keys[:-1]:
            tree = tree.setdefault(k, {})

        value = parse_conf_data(value, tomlfy=tomlfy, box_settings=self)
        tree[split_keys[-1]] = value

        if existing_data:
            new_data = object_merge(
                old=DynaBox({split_keys[0]: existing_data}),
                new=new_data,
                full_path=split_keys,
            )
        self.update(data=new_data, tomlfy=tomlfy, **kwargs)

    def set(
        self,
        key,
        value,
        loader_identifier=None,
        tomlfy=False,
        dotted_lookup=True,
        is_secret=False,
        merge=False,
    ):
        """Set a value storing references for the loader

        :param key: The key to store
        :param value: The value to store
        :param loader_identifier: Optional loader name e.g: toml, yaml etc.
        :param tomlfy: Bool define if value is parsed by toml (defaults False)
        :param is_secret: Bool define if secret values is hidden on logs.
        :param merge: Bool define if existing nested data will be merged.
        """
        nested_sep = self.get("NESTED_SEPARATOR_FOR_DYNACONF")
        if nested_sep and nested_sep in key:
            # turn FOO__bar__ZAZ in `FOO.bar.ZAZ`
            key = key.replace(nested_sep, ".")

        if "." in key and dotted_lookup is True:
            return self._dotted_set(
                key, value, loader_identifier=loader_identifier, tomlfy=tomlfy
            )

        value = parse_conf_data(value, tomlfy=tomlfy, box_settings=self)
        key = upperfy(key.strip())
        existing = getattr(self, key, None)

        if getattr(value, "_dynaconf_del", None):
            # just in case someone use a `@del` in a first level var.
            self.unset(key, force=True)
            return

        if getattr(value, "_dynaconf_reset", False):  # pragma: no cover
            # just in case someone use a `@reset` in a first level var.
            # NOTE: @reset/Reset is deprecated in v3.0.0
            value = value.unwrap()

        if getattr(value, "_dynaconf_merge", False):
            # just in case someone use a `@merge` in a first level var
            if existing:
                value = object_merge(existing, value.unwrap())
            else:
                value = value.unwrap()

        if existing is not None and existing != value:
            # `dynaconf_merge` used in file root `merge=True`
            if merge:
                value = object_merge(existing, value)
            else:
                # `dynaconf_merge` may be used within the key structure
                value = self._merge_before_set(key, existing, value, is_secret)

        if isinstance(value, dict):
            value = DynaBox(value, box_settings=self)

        self.store[key] = value
        self._deleted.discard(key)
        super(Settings, self).__setattr__(key, value)

        # set loader identifiers so cleaners know which keys to clean
        if loader_identifier and loader_identifier in self.loaded_by_loaders:
            self.loaded_by_loaders[loader_identifier][key] = value
        elif loader_identifier:
            self.loaded_by_loaders[loader_identifier] = {key: value}
        elif loader_identifier is None:
            # if .set is called without loader identifier it becomes
            # a default value and goes away only when explicitly unset
            self._defaults[key] = value

    def update(
        self,
        data=None,
        loader_identifier=None,
        tomlfy=False,
        is_secret=False,
        merge=False,
        **kwargs,
    ):
        """
        Update values in the current settings object without saving in stores::

            >>> from dynaconf import settings
            >>> print settings.NAME
            'Bruno'
            >>> settings.update({'NAME': 'John'}, other_value=1)
            >>> print settings.NAME
            'John'
            >>> print settings.OTHER_VALUE
            1

        :param data: Data to be updated
        :param loader_identifier: Only to be used by custom loaders
        :param tomlfy: Bool define if value is parsed by toml (defaults False)
        :param merge: Bool define if existing nested data will be merged.
        :param kwargs: extra values to update
        :return: None
        """
        data = data or {}
        data.update(kwargs)
        for key, value in data.items():
            self.set(
                key,
                value,
                loader_identifier=loader_identifier,
                tomlfy=tomlfy,
                is_secret=is_secret,
                merge=merge,
            )

    def _merge_before_set(self, key, existing, value, is_secret):
        """Merge the new value being set with the existing value before set"""
        global_merge = getattr(self, "MERGE_ENABLED_FOR_DYNACONF", False)
        if isinstance(value, dict):
            local_merge = value.pop(
                "dynaconf_merge", value.pop("dynaconf_merge_unique", None)
            )
            if local_merge not in (True, False, None) and not value:
                # In case `dynaconf_merge:` holds value not boolean - ref #241
                value = local_merge

            if global_merge or local_merge:
                value = object_merge(existing, value)

        if isinstance(value, (list, tuple)):
            local_merge = (
                "dynaconf_merge" in value or "dynaconf_merge_unique" in value
            )
            if global_merge or local_merge:
                value = list(value)
                unique = False
                if local_merge:
                    try:
                        value.remove("dynaconf_merge")
                    except ValueError:  # EAFP
                        value.remove("dynaconf_merge_unique")
                        unique = True
                value = object_merge(existing, value, unique=unique)
        return value

    @property
    def loaders(self):  # pragma: no cover
        """Return available loaders"""
        if self.LOADERS_FOR_DYNACONF in (None, 0, "0", "false", False):
            return []

        if not self._loaders:
            for loader_module_name in self.LOADERS_FOR_DYNACONF:
                loader = importlib.import_module(loader_module_name)
                self._loaders.append(loader)

        return self._loaders

    def reload(self, env=None, silent=None):  # pragma: no cover
        """Clean end Execute all loaders"""
        self.clean()
        self.execute_loaders(env, silent)

    def execute_loaders(self, env=None, silent=None, key=None, filename=None):
        """Execute all internal and registered loaders

        :param env: The environment to load
        :param silent: If loading erros is silenced
        :param key: if provided load a single key
        :param filename: optional custom filename to load
        """
        if key is None:
            default_loader(self, self._defaults)
        env = (env or self.current_env).upper()
        silent = silent or self.SILENT_ERRORS_FOR_DYNACONF
        self.pre_load(env, silent=silent, key=key)
        settings_loader(
            self, env=env, silent=silent, key=key, filename=filename
        )
        self.load_extra_yaml(env, silent, key)  # DEPRECATED
        enable_external_loaders(self)
        for loader in self.loaders:
            loader.load(self, env, silent=silent, key=key)
        self.load_includes(env, silent=silent, key=key)

    def pre_load(self, env, silent, key):
        """Do we have any file to pre-load before main settings file?"""
        preloads = self.get("PRELOAD_FOR_DYNACONF", [])
        if preloads:
            self.load_file(path=preloads, env=env, silent=silent, key=key)

    def load_includes(self, env, silent, key):
        """Do we have any nested includes we need to process?"""
        includes = self.get("DYNACONF_INCLUDE", [])
        includes.extend(ensure_a_list(self.get("INCLUDES_FOR_DYNACONF")))
        if includes:
            self.load_file(path=includes, env=env, silent=silent, key=key)
            # ensure env vars are the last thing loaded after all includes
            last_loader = self.loaders and self.loaders[-1]
            if last_loader and last_loader == env_loader:
                last_loader.load(self, env, silent, key)

    def load_file(self, path=None, env=None, silent=True, key=None):
        """Programmatically load files from ``path``.

        :param path: A single filename or a file list
        :param env: Which env to load from file (default current_env)
        :param silent: Should raise errors?
        :param key: Load a single key?
        """
        env = (env or self.current_env).upper()
        files = ensure_a_list(path)
        if files:
            already_loaded = set()
            for _filename in files:

                if py_loader.try_to_load_from_py_module_name(
                    obj=self, name=_filename, silent=True
                ):
                    # if it was possible to load from module name
                    # continue the loop.
                    continue

                filepath = os.path.join(
                    self._root_path or os.getcwd(), _filename
                )
                paths = [
                    p
                    for p in sorted(glob.glob(filepath))
                    if ".local." not in p
                ]
                local_paths = [
                    p for p in sorted(glob.glob(filepath)) if ".local." in p
                ]
                # Handle possible *.globs sorted alphanumeric
                for path in paths + local_paths:
                    if path in already_loaded:  # pragma: no cover
                        continue
                    settings_loader(
                        obj=self,
                        env=env,
                        silent=silent,
                        key=key,
                        filename=path,
                    )
                    already_loaded.add(path)

    @property
    def _root_path(self):
        """ROOT_PATH_FOR_DYNACONF or the path of first loaded file or '.'"""
        if self.ROOT_PATH_FOR_DYNACONF is not None:
            return self.ROOT_PATH_FOR_DYNACONF
        elif self._loaded_files:  # called once
            root_path = os.path.dirname(self._loaded_files[0])
            self.set("ROOT_PATH_FOR_DYNACONF", root_path)
            return root_path

    def load_extra_yaml(self, env, silent, key):
        """This is deprecated, kept for compat

        .. deprecated:: 1.0.0
            Use multiple settings or INCLUDES_FOR_DYNACONF files instead.
        """
        if self.get("YAML") is not None:
            warnings.warn(
                "The use of YAML var is deprecated, please define multiple "
                "filepaths instead: "
                "e.g: SETTINGS_FILE_FOR_DYNACONF = "
                "'settings.py,settings.yaml,settings.toml' or "
                "INCLUDES_FOR_DYNACONF=['path.toml', 'folder/*']"
            )
            yaml_loader.load(
                self,
                env=env,
                filename=self.find_file(self.get("YAML")),
                silent=silent,
                key=key,
            )

    def path_for(self, *args):
        """Path containing _root_path"""
        if args and args[0].startswith(os.path.sep):
            return os.path.join(*args)
        return os.path.join(self._root_path or os.getcwd(), *args)

    def find_file(self, *args, **kwargs):
        kwargs.setdefault("project_root", self._root_path)
        kwargs.setdefault(
            "skip_files", self.get("SKIP_FILES_FOR_DYNACONF", [])
        )
        return find_file(*args, **kwargs)

    def flag(self, key, env=None):
        """Feature flagging system
        write flags to redis
        $ dynaconf write redis -s DASHBOARD=1 -e premiumuser
        meaning: Any premium user has DASHBOARD feature enabled

        In your program do::

            # premium user has access to dashboard?
            >>> if settings.flag('dashboard', 'premiumuser'):
            ...     activate_dashboard()

        The value is ensured to be loaded fresh from redis server

        It also works with file settings but the recommended is redis
        as the data can be loaded once it is updated.

        :param key: The flag name
        :param env: The env to look for
        """
        env = env or self.ENVVAR_PREFIX_FOR_DYNACONF or "DYNACONF"
        with self.using_env(env):
            value = self.get_fresh(key)
            return value is True or value in true_values

    def populate_obj(self, obj, keys=None, ignore=None):
        """Given the `obj` populate it using self.store items.

        :param obj: An object to be populated, a class instance.
        :param keys: A list of keys to be included.
        :param ignore: A list of keys to be excluded.
        """
        keys = keys or self.keys()
        for key in keys:
            key = upperfy(key)
            if ignore and key in ignore:
                continue
            value = self.get(key, empty)
            if value is not empty:
                setattr(obj, key, value)

    def dynaconf(self):  # pragma: no cover
        """A proxy to access internal methods and attributes

        Starting in 3.0.0 Dynaconf now allows first level lower case
        keys that are not reserved keyword, so this is a proxy to
        internal methods and attrs.
        """
        return  # TOBE IMPLEMENTED

    @property
    def logger(self):  # pragma: no cover
        """backwards compatibility with pre 3.0 loaders
        In dynaconf 3.0.0 logger and debug messages has been removed.
        """
        warnings.warn(
            "logger and DEBUG messages has been removed on dynaconf 3.0.0"
        )
        import logging  # noqa

        return logging.getLogger("dynaconf")

    def is_overridden(self, setting):
        """This is to provide Django DJDT support: issue 382"""
        return False


"""Attributes created on Settings before 3.0.0"""
RESERVED_ATTRS = (
    [
        item[0]
        for item in inspect.getmembers(LazySettings)
        if not item[0].startswith("__")
    ]
    + [
        item[0]
        for item in inspect.getmembers(Settings)
        if not item[0].startswith("__")
    ]
    + [
        "_defaults",
        "_deleted",
        "_env_cache",
        "_fresh",
        "_kwargs",
        "_loaded_by_loaders",
        "_loaded_envs",
        "_loaded_files",
        "_loaders",
        "_memoized",
        "_not_installed_warnings",
        "_store",
        "_warn_dynaconf_global_settings",
        "environ",
        "SETTINGS_MODULE",
        "validators",
    ]
)



_I='COMP_CWORD'
_H='COMP_WORDS'
_G='fish'
_F='zsh'
_E='bash'
_D='_'
_C=False
_B=None
_A=True
import copy,os,re
from collections import abc
from .core import Argument
from .core import MultiCommand
from .core import Option
from .parser import split_arg_string
from .types import Choice
from .utils import echo
WORDBREAK='='
COMPLETION_SCRIPT_BASH='\n%(complete_func)s() {\n    local IFS=$\'\n\'\n    COMPREPLY=( $( env COMP_WORDS="${COMP_WORDS[*]}" \\\n                   COMP_CWORD=$COMP_CWORD \\\n                   %(autocomplete_var)s=complete $1 ) )\n    return 0\n}\n\n%(complete_func)setup() {\n    local COMPLETION_OPTIONS=""\n    local BASH_VERSION_ARR=(${BASH_VERSION//./ })\n    # Only BASH version 4.4 and later have the nosort option.\n    if [ ${BASH_VERSION_ARR[0]} -gt 4 ] || ([ ${BASH_VERSION_ARR[0]} -eq 4 ] && [ ${BASH_VERSION_ARR[1]} -ge 4 ]); then\n        COMPLETION_OPTIONS="-o nosort"\n    fi\n\n    complete $COMPLETION_OPTIONS -F %(complete_func)s %(script_names)s\n}\n\n%(complete_func)setup\n'
COMPLETION_SCRIPT_ZSH='\n#compdef %(script_names)s\n\n%(complete_func)s() {\n    local -a completions\n    local -a completions_with_descriptions\n    local -a response\n    (( ! $+commands[%(script_names)s] )) && return 1\n\n    response=("${(@f)$( env COMP_WORDS="${words[*]}" \\\n                        COMP_CWORD=$((CURRENT-1)) \\\n                        %(autocomplete_var)s="complete_zsh" \\\n                        %(script_names)s )}")\n\n    for key descr in ${(kv)response}; do\n      if [[ "$descr" == "_" ]]; then\n          completions+=("$key")\n      else\n          completions_with_descriptions+=("$key":"$descr")\n      fi\n    done\n\n    if [ -n "$completions_with_descriptions" ]; then\n        _describe -V unsorted completions_with_descriptions -U\n    fi\n\n    if [ -n "$completions" ]; then\n        compadd -U -V unsorted -a completions\n    fi\n    compstate[insert]="automenu"\n}\n\ncompdef %(complete_func)s %(script_names)s\n'
COMPLETION_SCRIPT_FISH='complete --no-files --command %(script_names)s --arguments "(env %(autocomplete_var)s=complete_fish COMP_WORDS=(commandline -cp) COMP_CWORD=(commandline -t) %(script_names)s)"'
_completion_scripts={_E:COMPLETION_SCRIPT_BASH,_F:COMPLETION_SCRIPT_ZSH,_G:COMPLETION_SCRIPT_FISH}
_invalid_ident_char_re=re.compile('[^a-zA-Z0-9_]')
def get_completion_script(prog_name,complete_var,shell):A=prog_name;B=_invalid_ident_char_re.sub('',A.replace('-',_D));C=_completion_scripts.get(shell,COMPLETION_SCRIPT_BASH);return (C%{'complete_func':f"_{B}_completion",'script_names':A,'autocomplete_var':complete_var}).strip()+';'
def resolve_ctx(cli,prog_name,args):
	B=args;A=cli.make_context(prog_name,B,resilient_parsing=_A);B=A.protected_args+A.args
	while B:
		if isinstance(A.command,MultiCommand):
			if not A.command.chain:
				E,C,B=A.command.resolve_command(A,B)
				if C is _B:return A
				A=C.make_context(E,B,parent=A,resilient_parsing=_A);B=A.protected_args+A.args
			else:
				while B:
					E,C,B=A.command.resolve_command(A,B)
					if C is _B:return A
					D=C.make_context(E,B,parent=A,allow_extra_args=_A,allow_interspersed_args=_C,resilient_parsing=_A);B=D.args
				A=D;B=D.protected_args+D.args
		else:break
	return A
def start_of_option(param_str):A=param_str;return A and A[:1]=='-'
def is_incomplete_option(all_args,cmd_param):
	A=cmd_param
	if not isinstance(A,Option):return _C
	if A.is_flag:return _C
	B=_B
	for (D,C) in enumerate(reversed([A for A in all_args if A!=WORDBREAK])):
		if D+1>A.nargs:break
		if start_of_option(C):B=C
	return _A if B and B in A.opts else _C
def is_incomplete_argument(current_params,cmd_param):
	A=cmd_param
	if not isinstance(A,Argument):return _C
	B=current_params[A.name]
	if B is _B:return _A
	if A.nargs==-1:return _A
	if isinstance(B,abc.Iterable)and A.nargs>1 and len(B)<A.nargs:return _A
	return _C
def get_user_autocompletions(ctx,args,incomplete,cmd_param):
	C=incomplete;A=cmd_param;B=[]
	if isinstance(A.type,Choice):B=[(B,_B)for B in A.type.choices if str(B).startswith(C)]
	elif A.autocompletion is not _B:D=A.autocompletion(ctx=ctx,args=args,incomplete=C);B=[A if isinstance(A,tuple)else(A,_B)for A in D]
	return B
def get_visible_commands_starting_with(ctx,starts_with):
	A=ctx
	for B in A.command.list_commands(A):
		if B.startswith(starts_with):
			C=A.command.get_command(A,B)
			if not C.hidden:yield C
def add_subcommand_completions(ctx,incomplete,completions_out):
	C=completions_out;B=incomplete;A=ctx
	if isinstance(A.command,MultiCommand):C.extend([(C.name,C.get_short_help_str())for C in get_visible_commands_starting_with(A,B)])
	while A.parent is not _B:
		A=A.parent
		if isinstance(A.command,MultiCommand)and A.command.chain:D=[C for C in get_visible_commands_starting_with(A,B)if C.name not in A.protected_args];C.extend([(A.name,A.get_short_help_str())for A in D])
def get_choices(cli,prog_name,args,incomplete):
	B=incomplete;D=copy.deepcopy(args);C=resolve_ctx(cli,prog_name,args)
	if C is _B:return[]
	G='--'in D
	if start_of_option(B)and WORDBREAK in B:F=B.partition(WORDBREAK);D.append(F[0]);B=F[2]
	elif B==WORDBREAK:B=''
	E=[]
	if not G and start_of_option(B):
		for A in C.command.params:
			if isinstance(A,Option)and not A.hidden:H=[B for B in A.opts+A.secondary_opts if B not in D or A.multiple];E.extend([(C,A.help)for C in H if C.startswith(B)])
		return E
	for A in C.command.params:
		if is_incomplete_option(D,A):return get_user_autocompletions(C,D,B,A)
	for A in C.command.params:
		if is_incomplete_argument(C.params,A):return get_user_autocompletions(C,D,B,A)
	add_subcommand_completions(C,B,E);return sorted(E)
def do_complete(cli,prog_name,include_descriptions):
	B=split_arg_string(os.environ[_H]);C=int(os.environ[_I]);E=B[1:C]
	try:D=B[C]
	except IndexError:D=''
	for A in get_choices(cli,prog_name,E,D):
		echo(A[0])
		if include_descriptions:echo(A[1]if A[1]else _D)
	return _A
def do_complete_fish(cli,prog_name):
	B=split_arg_string(os.environ[_H]);C=os.environ[_I];D=B[1:]
	for A in get_choices(cli,prog_name,D,C):
		if A[1]:echo(f"{A[0]}\t{A[1]}")
		else:echo(A[0])
	return _A
def bashcomplete(cli,prog_name,complete_var,complete_instr):
	C=complete_instr;B=prog_name
	if _D in C:D,A=C.split(_D,1)
	else:D=C;A=_E
	if D=='source':echo(get_completion_script(B,complete_var,A));return _A
	elif D=='complete':
		if A==_G:return do_complete_fish(cli,B)
		elif A in{_E,_F}:return do_complete(cli,B,A==_F)
	return _C


from __future__ import absolute_import
_F='serializer is not opened'
_E='serializer is closed'
_D='typ'
_C=False
_B=True
_A=None
from .error import YAMLError
from .compat import nprint,DBG_NODE,dbg,string_types,nprintf
from .util import RegExp
from .events import StreamStartEvent,StreamEndEvent,MappingStartEvent,MappingEndEvent,SequenceStartEvent,SequenceEndEvent,AliasEvent,ScalarEvent,DocumentStartEvent,DocumentEndEvent
from .nodes import MappingNode,ScalarNode,SequenceNode
if _C:from typing import Any,Dict,Union,Text,Optional;from .compat import VersionType
__all__=['Serializer','SerializerError']
class SerializerError(YAMLError):0
class Serializer:
	ANCHOR_TEMPLATE='id%03d';ANCHOR_RE=RegExp('id(?!000$)\\d{3,}')
	def __init__(A,encoding=_A,explicit_start=_A,explicit_end=_A,version=_A,tags=_A,dumper=_A):
		B=version;A.dumper=dumper
		if A.dumper is not _A:A.dumper._serializer=A
		A.use_encoding=encoding;A.use_explicit_start=explicit_start;A.use_explicit_end=explicit_end
		if isinstance(B,string_types):A.use_version=tuple(map(int,B.split('.')))
		else:A.use_version=B
		A.use_tags=tags;A.serialized_nodes={};A.anchors={};A.last_anchor_id=0;A.closed=_A;A._templated_id=_A
	@property
	def emitter(self):
		A=self
		if hasattr(A.dumper,_D):return A.dumper.emitter
		return A.dumper._emitter
	@property
	def resolver(self):
		A=self
		if hasattr(A.dumper,_D):A.dumper.resolver
		return A.dumper._resolver
	def open(A):
		if A.closed is _A:A.emitter.emit(StreamStartEvent(encoding=A.use_encoding));A.closed=_C
		elif A.closed:raise SerializerError(_E)
		else:raise SerializerError('serializer is already opened')
	def close(A):
		if A.closed is _A:raise SerializerError(_F)
		elif not A.closed:A.emitter.emit(StreamEndEvent());A.closed=_B
	def serialize(A,node):
		B=node
		if dbg(DBG_NODE):nprint('Serializing nodes');B.dump()
		if A.closed is _A:raise SerializerError(_F)
		elif A.closed:raise SerializerError(_E)
		A.emitter.emit(DocumentStartEvent(explicit=A.use_explicit_start,version=A.use_version,tags=A.use_tags));A.anchor_node(B);A.serialize_node(B,_A,_A);A.emitter.emit(DocumentEndEvent(explicit=A.use_explicit_end));A.serialized_nodes={};A.anchors={};A.last_anchor_id=0
	def anchor_node(B,node):
		A=node
		if A in B.anchors:
			if B.anchors[A]is _A:B.anchors[A]=B.generate_anchor(A)
		else:
			C=_A
			try:
				if A.anchor.always_dump:C=A.anchor.value
			except:pass
			B.anchors[A]=C
			if isinstance(A,SequenceNode):
				for D in A.value:B.anchor_node(D)
			elif isinstance(A,MappingNode):
				for (E,F) in A.value:B.anchor_node(E);B.anchor_node(F)
	def generate_anchor(A,node):
		try:B=node.anchor.value
		except:B=_A
		if B is _A:A.last_anchor_id+=1;return A.ANCHOR_TEMPLATE%A.last_anchor_id
		return B
	def serialize_node(B,node,parent,index):
		F=index;A=node;G=B.anchors[A]
		if A in B.serialized_nodes:B.emitter.emit(AliasEvent(G))
		else:
			B.serialized_nodes[A]=_B;B.resolver.descend_resolver(parent,F)
			if isinstance(A,ScalarNode):K=B.resolver.resolve(ScalarNode,A.value,(_B,_C));L=B.resolver.resolve(ScalarNode,A.value,(_C,_B));E=A.tag==K,A.tag==L,A.tag.startswith('tag:yaml.org,2002:');B.emitter.emit(ScalarEvent(G,A.tag,E,A.value,style=A.style,comment=A.comment))
			elif isinstance(A,SequenceNode):
				E=A.tag==B.resolver.resolve(SequenceNode,A.value,_B);C=A.comment;D=_A;H=_A
				if A.flow_style is _B:
					if C:H=C[0]
				if C and len(C)>2:D=C[2]
				else:D=_A
				B.emitter.emit(SequenceStartEvent(G,A.tag,E,flow_style=A.flow_style,comment=A.comment));F=0
				for M in A.value:B.serialize_node(M,A,F);F+=1
				B.emitter.emit(SequenceEndEvent(comment=[H,D]))
			elif isinstance(A,MappingNode):
				E=A.tag==B.resolver.resolve(MappingNode,A.value,_B);C=A.comment;D=_A;I=_A
				if A.flow_style is _B:
					if C:I=C[0]
				if C and len(C)>2:D=C[2]
				B.emitter.emit(MappingStartEvent(G,A.tag,E,flow_style=A.flow_style,comment=A.comment,nr_items=len(A.value)))
				for (J,N) in A.value:B.serialize_node(J,A,_A);B.serialize_node(N,A,J)
				B.emitter.emit(MappingEndEvent(comment=[I,D]))
			B.resolver.ascend_resolver()
def templated_id(s):return Serializer.ANCHOR_RE.match(s)


_E='replace'
_D=False
_C='\n'
_B='\r\n'
_A=None
import contextlib,io,os,shlex,shutil,sys,tempfile
from .  import formatting,termui,utils
from ._compat import _find_binary_reader
class EchoingStdin:
	def __init__(A,input,output):A._input=input;A._output=output
	def __getattr__(A,x):return getattr(A._input,x)
	def _echo(A,rv):A._output.write(rv);return rv
	def read(A,n=-1):return A._echo(A._input.read(n))
	def readline(A,n=-1):return A._echo(A._input.readline(n))
	def readlines(A):return[A._echo(B)for B in A._input.readlines()]
	def __iter__(A):return iter((A._echo(B)for B in A._input))
	def __repr__(A):return repr(A._input)
def make_input_stream(input,charset):
	if hasattr(input,'read'):
		A=_find_binary_reader(input)
		if A is not _A:return A
		raise TypeError('Could not find binary reader for input stream.')
	if input is _A:input=b''
	elif not isinstance(input,bytes):input=input.encode(charset)
	return io.BytesIO(input)
class Result:
	def __init__(A,runner,stdout_bytes,stderr_bytes,exit_code,exception,exc_info=_A):A.runner=runner;A.stdout_bytes=stdout_bytes;A.stderr_bytes=stderr_bytes;A.exit_code=exit_code;A.exception=exception;A.exc_info=exc_info
	@property
	def output(self):return self.stdout
	@property
	def stdout(self):return self.stdout_bytes.decode(self.runner.charset,_E).replace(_B,_C)
	@property
	def stderr(self):
		A=self
		if A.stderr_bytes is _A:raise ValueError('stderr not separately captured')
		return A.stderr_bytes.decode(A.runner.charset,_E).replace(_B,_C)
	def __repr__(A):B=repr(A.exception)if A.exception else'okay';return f"<{type(A).__name__} {B}>"
class CliRunner:
	def __init__(A,charset='utf-8',env=_A,echo_stdin=_D,mix_stderr=True):A.charset=charset;A.env=env or{};A.echo_stdin=echo_stdin;A.mix_stderr=mix_stderr
	def get_default_prog_name(A,cli):return cli.name or'root'
	def make_env(C,overrides=_A):
		A=overrides;B=dict(C.env)
		if A:B.update(A)
		return B
	@contextlib.contextmanager
	def isolation(self,input=_A,env=_A,color=_D):
		D=env;A=self;input=make_input_stream(input,A.charset);H=sys.stdin;I=sys.stdout;J=sys.stderr;K=formatting.FORCED_WIDTH;formatting.FORCED_WIDTH=80;D=A.make_env(D);E=io.BytesIO()
		if A.echo_stdin:input=EchoingStdin(input,E)
		input=io.TextIOWrapper(input,encoding=A.charset);sys.stdout=io.TextIOWrapper(E,encoding=A.charset)
		if not A.mix_stderr:F=io.BytesIO();sys.stderr=io.TextIOWrapper(F,encoding=A.charset)
		if A.mix_stderr:sys.stderr=sys.stdout
		sys.stdin=input
		def L(prompt=_A):sys.stdout.write(prompt or'');A=input.readline().rstrip(_B);sys.stdout.write(f"{A}\n");sys.stdout.flush();return A
		def M(prompt=_A):sys.stdout.write(f"{prompt or''}\n");sys.stdout.flush();return input.readline().rstrip(_B)
		def N(echo):
			A=sys.stdin.read(1)
			if echo:sys.stdout.write(A);sys.stdout.flush()
			return A
		O=color
		def P(stream=_A,color=_A):
			A=color
			if A is _A:return not O
			return not A
		Q=termui.visible_prompt_func;R=termui.hidden_prompt_func;S=termui._getchar;T=utils.should_strip_ansi;termui.visible_prompt_func=L;termui.hidden_prompt_func=M;termui._getchar=N;utils.should_strip_ansi=P;G={}
		try:
			for (B,C) in D.items():
				G[B]=os.environ.get(B)
				if C is _A:
					try:del os.environ[B]
					except Exception:pass
				else:os.environ[B]=C
			yield(E,not A.mix_stderr and F)
		finally:
			for (B,C) in G.items():
				if C is _A:
					try:del os.environ[B]
					except Exception:pass
				else:os.environ[B]=C
			sys.stdout=I;sys.stderr=J;sys.stdin=H;termui.visible_prompt_func=Q;termui.hidden_prompt_func=R;termui._getchar=S;utils.should_strip_ansi=T;formatting.FORCED_WIDTH=K
	def invoke(B,cli,args=_A,input=_A,env=_A,catch_exceptions=True,color=_D,**G):
		C=args;E=_A
		with B.isolation(input=input,env=env,color=color)as H:
			F=_A;A=0
			if isinstance(C,str):C=shlex.split(C)
			try:I=G.pop('prog_name')
			except KeyError:I=B.get_default_prog_name(cli)
			try:cli.main(args=C or(),prog_name=I,**G)
			except SystemExit as D:
				E=sys.exc_info();A=D.code
				if A is _A:A=0
				if A!=0:F=D
				if not isinstance(A,int):sys.stdout.write(str(A));sys.stdout.write(_C);A=1
			except Exception as D:
				if not catch_exceptions:raise
				F=D;A=1;E=sys.exc_info()
			finally:
				sys.stdout.flush();K=H[0].getvalue()
				if B.mix_stderr:J=_A
				else:J=H[1].getvalue()
		return Result(runner=B,stdout_bytes=K,stderr_bytes=J,exit_code=A,exception=F,exc_info=E)
	@contextlib.contextmanager
	def isolated_filesystem(self):
		B=os.getcwd();A=tempfile.mkdtemp();os.chdir(A)
		try:yield A
		finally:
			os.chdir(B)
			try:shutil.rmtree(A)
			except OSError:pass


_I='error'
_H='original'
_G='value'
_F='key'
_E='Binding'
_D='line'
_C='string'
_B='Original'
_A=None
import codecs,re
from .compat import IS_TYPE_CHECKING,to_text
if IS_TYPE_CHECKING:from typing import IO,Iterator,Match,NamedTuple,Optional,Pattern,Sequence,Text,Tuple
def make_regex(string,extra_flags=0):return re.compile(to_text(string),re.UNICODE|extra_flags)
_newline=make_regex('(\\r\\n|\\n|\\r)')
_multiline_whitespace=make_regex('\\s*',extra_flags=re.MULTILINE)
_whitespace=make_regex('[^\\S\\r\\n]*')
_export=make_regex('(?:export[^\\S\\r\\n]+)?')
_single_quoted_key=make_regex("'([^']+)'")
_unquoted_key=make_regex('([^=\\#\\s]+)')
_equal_sign=make_regex('(=[^\\S\\r\\n]*)')
_single_quoted_value=make_regex("'((?:\\\\'|[^'])*)'")
_double_quoted_value=make_regex('"((?:\\\\"|[^"])*)"')
_unquoted_value_part=make_regex('([^ \\r\\n]*)')
_comment=make_regex('(?:[^\\S\\r\\n]*#[^\\r\\n]*)?')
_end_of_line=make_regex('[^\\S\\r\\n]*(?:\\r\\n|\\n|\\r|$)')
_rest_of_line=make_regex('[^\\r\\n]*(?:\\r|\\n|\\r\\n)?')
_double_quote_escapes=make_regex('\\\\[\\\\\'\\"abfnrtv]')
_single_quote_escapes=make_regex("\\\\[\\\\']")
try:import typing;Original=typing.NamedTuple(_B,[(_C,typing.Text),(_D,int)]);Binding=typing.NamedTuple(_E,[(_F,typing.Optional[typing.Text]),(_G,typing.Optional[typing.Text]),(_H,Original),(_I,bool)])
except ImportError:from collections import namedtuple;Original=namedtuple(_B,[_C,_D]);Binding=namedtuple(_E,[_F,_G,_H,_I])
class Position:
	def __init__(A,chars,line):A.chars=chars;A.line=line
	@classmethod
	def start(A):return A(chars=0,line=1)
	def set(A,other):B=other;A.chars=B.chars;A.line=B.line
	def advance(A,string):B=string;A.chars+=len(B);A.line+=len(re.findall(_newline,B))
class Error(Exception):0
class Reader:
	def __init__(A,stream):A.string=stream.read();A.position=Position.start();A.mark=Position.start()
	def has_next(A):return A.position.chars<len(A.string)
	def set_mark(A):A.mark.set(A.position)
	def get_marked(A):return Original(string=A.string[A.mark.chars:A.position.chars],line=A.mark.line)
	def peek(A,count):return A.string[A.position.chars:A.position.chars+count]
	def read(A,count):
		C=count;B=A.string[A.position.chars:A.position.chars+C]
		if len(B)<C:raise Error('read: End of string')
		A.position.advance(B);return B
	def read_regex(A,regex):
		B=regex.match(A.string,A.position.chars)
		if B is _A:raise Error('read_regex: Pattern not found')
		A.position.advance(A.string[B.start():B.end()]);return B.groups()
def decode_escapes(regex,string):
	def A(match):return codecs.decode(match.group(0),'unicode-escape')
	return regex.sub(A,string)
def parse_key(reader):
	A=reader;B=A.peek(1)
	if B=='#':return _A
	elif B=="'":C,=A.read_regex(_single_quoted_key)
	else:C,=A.read_regex(_unquoted_key)
	return C
def parse_unquoted_value(reader):
	A=reader;B=''
	while True:
		D,=A.read_regex(_unquoted_value_part);B+=D;C=A.peek(2)
		if len(C)<2 or C[0]in'\r\n'or C[1]in' #\r\n':return B
		B+=A.read(2)
def parse_value(reader):
	A=reader;B=A.peek(1)
	if B=="'":C,=A.read_regex(_single_quoted_value);return decode_escapes(_single_quote_escapes,C)
	elif B=='"':C,=A.read_regex(_double_quoted_value);return decode_escapes(_double_quote_escapes,C)
	elif B in('','\n','\r'):return''
	else:return parse_unquoted_value(A)
def parse_binding(reader):
	D=False;A=reader;A.set_mark()
	try:
		A.read_regex(_multiline_whitespace)
		if not A.has_next():return Binding(key=_A,value=_A,original=A.get_marked(),error=D)
		A.read_regex(_export);C=parse_key(A);A.read_regex(_whitespace)
		if A.peek(1)=='=':A.read_regex(_equal_sign);B=parse_value(A)
		else:B=_A
		A.read_regex(_comment);A.read_regex(_end_of_line);return Binding(key=C,value=B,original=A.get_marked(),error=D)
	except Error:A.read_regex(_rest_of_line);return Binding(key=_A,value=_A,original=A.get_marked(),error=True)
def parse_stream(stream):
	A=Reader(stream)
	while A.has_next():yield parse_binding(A)


_H='toml'
_G='box_dots'
_F='BoxList is frozen'
_E='frozen_box'
_D=False
_C='strict'
_B='utf-8'
_A=None
import copy,re
from typing import Iterable
from dynaconf.vendor import box
from .converters import _to_yaml,_from_yaml,_to_json,_from_json,_to_toml,_from_toml,_to_csv,_from_csv,BOX_PARAMETERS
from .exceptions import BoxError,BoxTypeError,BoxKeyError
_list_pos_re=re.compile('\\[(\\d+)\\]')
class BoxList(list):
	def __init__(A,iterable=_A,box_class=box.Box,**C):
		B=iterable;A.box_class=box_class;A.box_options=C;A.box_org_ref=A.box_org_ref=id(B)if B else 0
		if B:
			for D in B:A.append(D)
		if C.get(_E):
			def E(*A,**B):raise BoxError(_F)
			for F in ['append','extend','insert','pop','remove','reverse','sort']:A.__setattr__(F,E)
	def __getitem__(B,item):
		A=item
		if B.box_options.get(_G)and isinstance(A,str)and A.startswith('['):
			C=_list_pos_re.search(A);D=super(BoxList,B).__getitem__(int(C.groups()[0]))
			if len(C.group())==len(A):return D
			return D.__getitem__(A[len(C.group()):].lstrip('.'))
		return super(BoxList,B).__getitem__(A)
	def __delitem__(A,key):
		if A.box_options.get(_E):raise BoxError(_F)
		super(BoxList,A).__delitem__(key)
	def __setitem__(B,key,value):
		C=value;A=key
		if B.box_options.get(_E):raise BoxError(_F)
		if B.box_options.get(_G)and isinstance(A,str)and A.startswith('['):
			D=_list_pos_re.search(A);E=int(D.groups()[0])
			if len(D.group())==len(A):return super(BoxList,B).__setitem__(E,C)
			return super(BoxList,B).__getitem__(E).__setitem__(A[len(D.group()):].lstrip('.'),C)
		super(BoxList,B).__setitem__(A,C)
	def _is_intact_type(A,obj):
		C='box_intact_types'
		try:
			if A.box_options.get(C)and isinstance(obj,A.box_options[C]):return True
		except AttributeError as B:
			if'box_options'in A.__dict__:raise BoxKeyError(B)
		return _D
	def append(A,p_object):
		B=p_object
		if isinstance(B,dict)and not A._is_intact_type(B):
			try:B=A.box_class(B,**A.box_options)
			except AttributeError as C:
				if'box_class'in A.__dict__:raise BoxKeyError(C)
		elif isinstance(B,list)and not A._is_intact_type(B):
			try:B=A if id(B)==A.box_org_ref else BoxList(B,**A.box_options)
			except AttributeError as C:
				if'box_org_ref'in A.__dict__:raise BoxKeyError(C)
		super(BoxList,A).append(B)
	def extend(A,iterable):
		for B in iterable:A.append(B)
	def insert(B,index,p_object):
		A=p_object
		if isinstance(A,dict)and not B._is_intact_type(A):A=B.box_class(A,**B.box_options)
		elif isinstance(A,list)and not B._is_intact_type(A):A=B if id(A)==B.box_org_ref else BoxList(A)
		super(BoxList,B).insert(index,A)
	def __repr__(A):return f"<BoxList: {A.to_list()}>"
	def __str__(A):return str(A.to_list())
	def __copy__(A):return BoxList((B for B in A),A.box_class,**A.box_options)
	def __deepcopy__(B,memo=_A):
		A=memo;C=B.__class__();A=A or{};A[id(B)]=C
		for D in B:C.append(copy.deepcopy(D,memo=A))
		return C
	def __hash__(A):
		if A.box_options.get(_E):B=98765;B^=hash(tuple(A));return B
		raise BoxTypeError("unhashable type: 'BoxList'")
	def to_list(C):
		A=[]
		for B in C:
			if B is C:A.append(A)
			elif isinstance(B,box.Box):A.append(B.to_dict())
			elif isinstance(B,BoxList):A.append(B.to_list())
			else:A.append(B)
		return A
	def to_json(D,filename=_A,encoding=_B,errors=_C,multiline=_D,**E):
		C=errors;B=encoding;A=filename
		if A and multiline:
			F=[_to_json(A,filename=_D,encoding=B,errors=C,**E)for A in D]
			with open(A,'w',encoding=B,errors=C)as G:G.write('\n'.join(F))
		else:return _to_json(D.to_list(),filename=A,encoding=B,errors=C,**E)
	@classmethod
	def from_json(E,json_string=_A,filename=_A,encoding=_B,errors=_C,multiline=_D,**A):
		D={}
		for B in list(A.keys()):
			if B in BOX_PARAMETERS:D[B]=A.pop(B)
		C=_from_json(json_string,filename=filename,encoding=encoding,errors=errors,multiline=multiline,**A)
		if not isinstance(C,list):raise BoxError(f"json data not returned as a list, but rather a {type(C).__name__}")
		return E(C,**D)
	def to_yaml(A,filename=_A,default_flow_style=_D,encoding=_B,errors=_C,**B):return _to_yaml(A.to_list(),filename=filename,default_flow_style=default_flow_style,encoding=encoding,errors=errors,**B)
	@classmethod
	def from_yaml(E,yaml_string=_A,filename=_A,encoding=_B,errors=_C,**A):
		D={}
		for B in list(A.keys()):
			if B in BOX_PARAMETERS:D[B]=A.pop(B)
		C=_from_yaml(yaml_string=yaml_string,filename=filename,encoding=encoding,errors=errors,**A)
		if not isinstance(C,list):raise BoxError(f"yaml data not returned as a list but rather a {type(C).__name__}")
		return E(C,**D)
	def to_toml(A,filename=_A,key_name=_H,encoding=_B,errors=_C):return _to_toml({key_name:A.to_list()},filename=filename,encoding=encoding,errors=errors)
	@classmethod
	def from_toml(F,toml_string=_A,filename=_A,key_name=_H,encoding=_B,errors=_C,**C):
		A=key_name;D={}
		for B in list(C.keys()):
			if B in BOX_PARAMETERS:D[B]=C.pop(B)
		E=_from_toml(toml_string=toml_string,filename=filename,encoding=encoding,errors=errors)
		if A not in E:raise BoxError(f"{A} was not found.")
		return F(E[A],**D)
	def to_csv(A,filename,encoding=_B,errors=_C):_to_csv(A,filename=filename,encoding=encoding,errors=errors)
	@classmethod
	def from_csv(A,filename,encoding=_B,errors=_C):return A(_from_csv(filename=filename,encoding=encoding,errors=errors))


# pragma: no cover
"""
Implement basic assertions to be used in assertion action
"""


def eq(value, other):
    """Equal"""
    return value == other


def ne(value, other):
    """Not equal"""
    return value != other


def gt(value, other):
    """Greater than"""
    return value > other


def lt(value, other):
    """Lower than"""
    return value < other


def gte(value, other):
    """Greater than or equal"""
    return value >= other


def lte(value, other):
    """Lower than or equal"""
    return value <= other


def identity(value, other):
    """Identity check using ID"""
    return value is other


def is_type_of(value, other):
    """Type check"""
    return isinstance(value, other)


def is_in(value, other):
    """Existence"""
    return value in other


def is_not_in(value, other):
    """Inexistence"""
    return value not in other


def cont(value, other):
    """Contains"""
    return other in value


def len_eq(value, other):
    """Length Equal"""
    return len(value) == other


def len_ne(value, other):
    """Length Not equal"""
    return len(value) != other


def len_min(value, other):
    """Minimum length"""
    return len(value) >= other


def len_max(value, other):
    """Maximum lenght"""
    return len(value) <= other


def startswith(value, term):
    """returns value.startswith(term) result"""
    return value.startswith(term)



_F='text'
_E='replace'
_D='utf-8'
_C=True
_B=False
_A=None
import os,stat
from datetime import datetime
from ._compat import _get_argv_encoding
from ._compat import filename_to_ui
from ._compat import get_filesystem_encoding
from ._compat import get_strerror
from ._compat import open_stream
from .exceptions import BadParameter
from .utils import LazyFile
from .utils import safecall
class ParamType:
	is_composite=_B;name=_A;envvar_list_splitter=_A
	def __call__(B,value,param=_A,ctx=_A):
		A=value
		if A is not _A:return B.convert(A,param,ctx)
	def get_metavar(A,param):0
	def get_missing_message(A,param):0
	def convert(A,value,param,ctx):return value
	def split_envvar_value(A,rv):return (rv or'').split(A.envvar_list_splitter)
	def fail(A,message,param=_A,ctx=_A):raise BadParameter(message,ctx=ctx,param=param)
class CompositeParamType(ParamType):
	is_composite=_C
	@property
	def arity(self):raise NotImplementedError()
class FuncParamType(ParamType):
	def __init__(A,func):A.name=func.__name__;A.func=func
	def convert(B,value,param,ctx):
		A=value
		try:return B.func(A)
		except ValueError:
			try:A=str(A)
			except UnicodeError:A=A.decode(_D,_E)
			B.fail(A,param,ctx)
class UnprocessedParamType(ParamType):
	name=_F
	def convert(A,value,param,ctx):return value
	def __repr__(A):return'UNPROCESSED'
class StringParamType(ParamType):
	name=_F
	def convert(D,value,param,ctx):
		A=value
		if isinstance(A,bytes):
			B=_get_argv_encoding()
			try:A=A.decode(B)
			except UnicodeError:
				C=get_filesystem_encoding()
				if C!=B:
					try:A=A.decode(C)
					except UnicodeError:A=A.decode(_D,_E)
				else:A=A.decode(_D,_E)
			return A
		return A
	def __repr__(A):return'STRING'
class Choice(ParamType):
	name='choice'
	def __init__(A,choices,case_sensitive=_C):A.choices=choices;A.case_sensitive=case_sensitive
	def get_metavar(A,param):return f"[{'|'.join(A.choices)}]"
	def get_missing_message(A,param):B=',\n\t'.join(A.choices);return f"Choose from:\n\t{B}"
	def convert(D,value,param,ctx):
		E=value;B=ctx;C=E;A={A:A for A in D.choices}
		if B is not _A and B.token_normalize_func is not _A:C=B.token_normalize_func(E);A={B.token_normalize_func(C):D for(C,D)in A.items()}
		if not D.case_sensitive:C=C.casefold();A={B.casefold():C for(B,C)in A.items()}
		if C in A:return A[C]
		D.fail(f"invalid choice: {E}. (choose from {', '.join(D.choices)})",param,B)
	def __repr__(A):return f"Choice({list(A.choices)})"
class DateTime(ParamType):
	name='datetime'
	def __init__(A,formats=_A):A.formats=formats or['%Y-%m-%d','%Y-%m-%dT%H:%M:%S','%Y-%m-%d %H:%M:%S']
	def get_metavar(A,param):return f"[{'|'.join(A.formats)}]"
	def _try_to_convert_date(A,value,format):
		try:return datetime.strptime(value,format)
		except ValueError:return _A
	def convert(A,value,param,ctx):
		B=value
		for format in A.formats:
			C=A._try_to_convert_date(B,format)
			if C:return C
		A.fail(f"invalid datetime format: {B}. (choose from {', '.join(A.formats)})")
	def __repr__(A):return'DateTime'
class IntParamType(ParamType):
	name='integer'
	def convert(B,value,param,ctx):
		A=value
		try:return int(A)
		except ValueError:B.fail(f"{A} is not a valid integer",param,ctx)
	def __repr__(A):return'INT'
class IntRange(IntParamType):
	name='integer range'
	def __init__(A,min=_A,max=_A,clamp=_B):A.min=min;A.max=max;A.clamp=clamp
	def convert(A,value,param,ctx):
		D=ctx;C=param;B=IntParamType.convert(A,value,C,D)
		if A.clamp:
			if A.min is not _A and B<A.min:return A.min
			if A.max is not _A and B>A.max:return A.max
		if A.min is not _A and B<A.min or A.max is not _A and B>A.max:
			if A.min is _A:A.fail(f"{B} is bigger than the maximum valid value {A.max}.",C,D)
			elif A.max is _A:A.fail(f"{B} is smaller than the minimum valid value {A.min}.",C,D)
			else:A.fail(f"{B} is not in the valid range of {A.min} to {A.max}.",C,D)
		return B
	def __repr__(A):return f"IntRange({A.min}, {A.max})"
class FloatParamType(ParamType):
	name='float'
	def convert(B,value,param,ctx):
		A=value
		try:return float(A)
		except ValueError:B.fail(f"{A} is not a valid floating point value",param,ctx)
	def __repr__(A):return'FLOAT'
class FloatRange(FloatParamType):
	name='float range'
	def __init__(A,min=_A,max=_A,clamp=_B):A.min=min;A.max=max;A.clamp=clamp
	def convert(A,value,param,ctx):
		D=ctx;C=param;B=FloatParamType.convert(A,value,C,D)
		if A.clamp:
			if A.min is not _A and B<A.min:return A.min
			if A.max is not _A and B>A.max:return A.max
		if A.min is not _A and B<A.min or A.max is not _A and B>A.max:
			if A.min is _A:A.fail(f"{B} is bigger than the maximum valid value {A.max}.",C,D)
			elif A.max is _A:A.fail(f"{B} is smaller than the minimum valid value {A.min}.",C,D)
			else:A.fail(f"{B} is not in the valid range of {A.min} to {A.max}.",C,D)
		return B
	def __repr__(A):return f"FloatRange({A.min}, {A.max})"
class BoolParamType(ParamType):
	name='boolean'
	def convert(B,value,param,ctx):
		A=value
		if isinstance(A,bool):return bool(A)
		A=A.lower()
		if A in('true','t','1','yes','y'):return _C
		elif A in('false','f','0','no','n'):return _B
		B.fail(f"{A} is not a valid boolean",param,ctx)
	def __repr__(A):return'BOOL'
class UUIDParameterType(ParamType):
	name='uuid'
	def convert(B,value,param,ctx):
		A=value;import uuid
		try:return uuid.UUID(A)
		except ValueError:B.fail(f"{A} is not a valid UUID value",param,ctx)
	def __repr__(A):return'UUID'
class File(ParamType):
	name='filename';envvar_list_splitter=os.path.pathsep
	def __init__(A,mode='r',encoding=_A,errors='strict',lazy=_A,atomic=_B):A.mode=mode;A.encoding=encoding;A.errors=errors;A.lazy=lazy;A.atomic=atomic
	def resolve_lazy_flag(A,value):
		if A.lazy is not _A:return A.lazy
		if value=='-':return _B
		elif'w'in A.mode:return _C
		return _B
	def convert(A,value,param,ctx):
		C=ctx;B=value
		try:
			if hasattr(B,'read')or hasattr(B,'write'):return B
			E=A.resolve_lazy_flag(B)
			if E:
				D=LazyFile(B,A.mode,A.encoding,A.errors,atomic=A.atomic)
				if C is not _A:C.call_on_close(D.close_intelligently)
				return D
			D,F=open_stream(B,A.mode,A.encoding,A.errors,atomic=A.atomic)
			if C is not _A:
				if F:C.call_on_close(safecall(D.close))
				else:C.call_on_close(safecall(D.flush))
			return D
		except OSError as G:A.fail(f"Could not open file: {filename_to_ui(B)}: {get_strerror(G)}",param,C)
class Path(ParamType):
	envvar_list_splitter=os.path.pathsep
	def __init__(A,exists=_B,file_okay=_C,dir_okay=_C,writable=_B,readable=_C,resolve_path=_B,allow_dash=_B,path_type=_A):
		A.exists=exists;A.file_okay=file_okay;A.dir_okay=dir_okay;A.writable=writable;A.readable=readable;A.resolve_path=resolve_path;A.allow_dash=allow_dash;A.type=path_type
		if A.file_okay and not A.dir_okay:A.name='file';A.path_type='File'
		elif A.dir_okay and not A.file_okay:A.name='directory';A.path_type='Directory'
		else:A.name='path';A.path_type='Path'
	def coerce_path_result(B,rv):
		A=rv
		if B.type is not _A and not isinstance(A,B.type):
			if B.type is str:A=A.decode(get_filesystem_encoding())
			else:A=A.encode(get_filesystem_encoding())
		return A
	def convert(A,value,param,ctx):
		E=ctx;D=param;B=value;C=B;G=A.file_okay and A.allow_dash and C in(b'-','-')
		if not G:
			if A.resolve_path:C=os.path.realpath(C)
			try:F=os.stat(C)
			except OSError:
				if not A.exists:return A.coerce_path_result(C)
				A.fail(f"{A.path_type} {filename_to_ui(B)!r} does not exist.",D,E)
			if not A.file_okay and stat.S_ISREG(F.st_mode):A.fail(f"{A.path_type} {filename_to_ui(B)!r} is a file.",D,E)
			if not A.dir_okay and stat.S_ISDIR(F.st_mode):A.fail(f"{A.path_type} {filename_to_ui(B)!r} is a directory.",D,E)
			if A.writable and not os.access(B,os.W_OK):A.fail(f"{A.path_type} {filename_to_ui(B)!r} is not writable.",D,E)
			if A.readable and not os.access(B,os.R_OK):A.fail(f"{A.path_type} {filename_to_ui(B)!r} is not readable.",D,E)
		return A.coerce_path_result(C)
class Tuple(CompositeParamType):
	def __init__(A,types):A.types=[convert_type(A)for A in types]
	@property
	def name(self):return f"<{' '.join((A.name for A in self.types))}>"
	@property
	def arity(self):return len(self.types)
	def convert(A,value,param,ctx):
		B=value
		if len(B)!=len(A.types):raise TypeError('It would appear that nargs is set to conflict with the composite type arity.')
		return tuple((C(D,param,ctx)for(C,D)in zip(A.types,B)))
def convert_type(ty,default=_A):
	B=default;A=ty;C=_B
	if A is _A and B is not _A:
		if isinstance(B,tuple):A=tuple(map(type,B))
		else:A=type(B)
		C=_C
	if isinstance(A,tuple):return Tuple(A)
	if isinstance(A,ParamType):return A
	if A is str or A is _A:return STRING
	if A is int:return INT
	if A is bool and not C:return BOOL
	if A is float:return FLOAT
	if C:return STRING
	if __debug__:
		try:
			if issubclass(A,ParamType):raise AssertionError(f"Attempted to use an uninstantiated parameter type ({A}).")
		except TypeError:pass
	return FuncParamType(A)
UNPROCESSED=UnprocessedParamType()
STRING=StringParamType()
INT=IntParamType()
FLOAT=FloatParamType()
BOOL=BoolParamType()
UUID=UUIDParameterType()


_A=False
if _A:from typing import Any,Dict,Optional,List,Union,Optional,Iterator
anchor_attrib='_yaml_anchor'
class Anchor:
	__slots__='value','always_dump';attrib=anchor_attrib
	def __init__(A):A.value=None;A.always_dump=_A
	def __repr__(A):B=', (always dump)'if A.always_dump else'';return 'Anchor({!r}{})'.format(A.value,B)


from __future__ import absolute_import
_F='\ufeff'
_E='\x00'
_D=False
_C='ascii'
_B='\n'
_A=None
import codecs
from .error import YAMLError,FileMark,StringMark,YAMLStreamError
from .compat import text_type,binary_type,PY3,UNICODE_SIZE
from .util import RegExp
if _D:from typing import Any,Dict,Optional,List,Union,Text,Tuple,Optional
__all__=['Reader','ReaderError']
class ReaderError(YAMLError):
	def __init__(A,name,position,character,encoding,reason):A.name=name;A.character=character;A.position=position;A.encoding=encoding;A.reason=reason
	def __str__(A):
		if isinstance(A.character,binary_type):return'\'%s\' codec can\'t decode byte #x%02x: %s\n  in "%s", position %d'%(A.encoding,ord(A.character),A.reason,A.name,A.position)
		else:return'unacceptable character #x%04x: %s\n  in "%s", position %d'%(A.character,A.reason,A.name,A.position)
class Reader:
	def __init__(A,stream,loader=_A):
		A.loader=loader
		if A.loader is not _A and getattr(A.loader,'_reader',_A)is _A:A.loader._reader=A
		A.reset_reader();A.stream=stream
	def reset_reader(A):A.name=_A;A.stream_pointer=0;A.eof=True;A.buffer='';A.pointer=0;A.raw_buffer=_A;A.raw_decode=_A;A.encoding=_A;A.index=0;A.line=0;A.column=0
	@property
	def stream(self):
		try:return self._stream
		except AttributeError:raise YAMLStreamError('input stream needs to specified')
	@stream.setter
	def stream(self,val):
		B=val;A=self
		if B is _A:return
		A._stream=_A
		if isinstance(B,text_type):A.name='<unicode string>';A.check_printable(B);A.buffer=B+_E
		elif isinstance(B,binary_type):A.name='<byte string>';A.raw_buffer=B;A.determine_encoding()
		else:
			if not hasattr(B,'read'):raise YAMLStreamError('stream argument needs to have a read() method')
			A._stream=B;A.name=getattr(A.stream,'name','<file>');A.eof=_D;A.raw_buffer=_A;A.determine_encoding()
	def peek(A,index=0):
		B=index
		try:return A.buffer[A.pointer+B]
		except IndexError:A.update(B+1);return A.buffer[A.pointer+B]
	def prefix(A,length=1):
		B=length
		if A.pointer+B>=len(A.buffer):A.update(B)
		return A.buffer[A.pointer:A.pointer+B]
	def forward_1_1(A,length=1):
		B=length
		if A.pointer+B+1>=len(A.buffer):A.update(B+1)
		while B!=0:
			C=A.buffer[A.pointer];A.pointer+=1;A.index+=1
			if C in'\n\x85\u2028\u2029'or C=='\r'and A.buffer[A.pointer]!=_B:A.line+=1;A.column=0
			elif C!=_F:A.column+=1
			B-=1
	def forward(A,length=1):
		B=length
		if A.pointer+B+1>=len(A.buffer):A.update(B+1)
		while B!=0:
			C=A.buffer[A.pointer];A.pointer+=1;A.index+=1
			if C==_B or C=='\r'and A.buffer[A.pointer]!=_B:A.line+=1;A.column=0
			elif C!=_F:A.column+=1
			B-=1
	def get_mark(A):
		if A.stream is _A:return StringMark(A.name,A.index,A.line,A.column,A.buffer,A.pointer)
		else:return FileMark(A.name,A.index,A.line,A.column)
	def determine_encoding(A):
		while not A.eof and(A.raw_buffer is _A or len(A.raw_buffer)<2):A.update_raw()
		if isinstance(A.raw_buffer,binary_type):
			if A.raw_buffer.startswith(codecs.BOM_UTF16_LE):A.raw_decode=codecs.utf_16_le_decode;A.encoding='utf-16-le'
			elif A.raw_buffer.startswith(codecs.BOM_UTF16_BE):A.raw_decode=codecs.utf_16_be_decode;A.encoding='utf-16-be'
			else:A.raw_decode=codecs.utf_8_decode;A.encoding='utf-8'
		A.update(1)
	if UNICODE_SIZE==2:NON_PRINTABLE=RegExp('[^\t\n\r -~\x85\xa0-\ud7ff\ue000-�]')
	else:NON_PRINTABLE=RegExp('[^\t\n\r -~\x85\xa0-\ud7ff\ue000-�𐀀-\U0010ffff]')
	_printable_ascii=('\t\n\r'+''.join(map(chr,range(32,127)))).encode(_C)
	@classmethod
	def _get_non_printable_ascii(D,data):
		A=data.encode(_C);B=A.translate(_A,D._printable_ascii)
		if not B:return _A
		C=B[:1];return A.index(C),C.decode(_C)
	@classmethod
	def _get_non_printable_regex(B,data):
		A=B.NON_PRINTABLE.search(data)
		if not bool(A):return _A
		return A.start(),A.group()
	@classmethod
	def _get_non_printable(A,data):
		try:return A._get_non_printable_ascii(data)
		except UnicodeEncodeError:return A._get_non_printable_regex(data)
	def check_printable(A,data):
		B=A._get_non_printable(data)
		if B is not _A:C,D=B;E=A.index+(len(A.buffer)-A.pointer)+C;raise ReaderError(A.name,E,ord(D),'unicode','special characters are not allowed')
	def update(A,length):
		if A.raw_buffer is _A:return
		A.buffer=A.buffer[A.pointer:];A.pointer=0
		while len(A.buffer)<length:
			if not A.eof:A.update_raw()
			if A.raw_decode is not _A:
				try:C,E=A.raw_decode(A.raw_buffer,'strict',A.eof)
				except UnicodeDecodeError as B:
					if PY3:F=A.raw_buffer[B.start]
					else:F=B.object[B.start]
					if A.stream is not _A:D=A.stream_pointer-len(A.raw_buffer)+B.start
					elif A.stream is not _A:D=A.stream_pointer-len(A.raw_buffer)+B.start
					else:D=B.start
					raise ReaderError(A.name,D,F,B.encoding,B.reason)
			else:C=A.raw_buffer;E=len(C)
			A.check_printable(C);A.buffer+=C;A.raw_buffer=A.raw_buffer[E:]
			if A.eof:A.buffer+=_E;A.raw_buffer=_A;break
	def update_raw(A,size=_A):
		C=size
		if C is _A:C=4096 if PY3 else 1024
		B=A.stream.read(C)
		if A.raw_buffer is _A:A.raw_buffer=B
		else:A.raw_buffer+=B
		A.stream_pointer+=len(B)
		if not B:A.eof=True


# docker run -e 'VAULT_DEV_ROOT_TOKEN_ID=myroot' -p 8200:8200 vault
# pip install hvac
from dynaconf.utils import build_env_list
from dynaconf.utils.parse_conf import parse_conf_data

try:
    import boto3
except ImportError:
    boto3 = None

try:
    from hvac import Client
    from hvac.exceptions import InvalidPath
except ImportError:
    raise ImportError(
        "vault package is not installed in your environment. "
        "`pip install dynaconf[vault]` or disable the vault loader with "
        "export VAULT_ENABLED_FOR_DYNACONF=false"
    )


IDENTIFIER = "vault"


# backwards compatibility
_get_env_list = build_env_list


def get_client(obj):
    client = Client(
        **{k: v for k, v in obj.VAULT_FOR_DYNACONF.items() if v is not None}
    )
    if obj.VAULT_ROLE_ID_FOR_DYNACONF is not None:
        client.auth_approle(
            role_id=obj.VAULT_ROLE_ID_FOR_DYNACONF,
            secret_id=obj.get("VAULT_SECRET_ID_FOR_DYNACONF"),
        )
    elif obj.VAULT_ROOT_TOKEN_FOR_DYNACONF is not None:
        client.token = obj.VAULT_ROOT_TOKEN_FOR_DYNACONF
    elif obj.VAULT_AUTH_WITH_IAM_FOR_DYNACONF:
        if boto3 is None:
            raise ImportError(
                "boto3 package is not installed in your environment. "
                "`pip install boto3` or disable the VAULT_AUTH_WITH_IAM"
            )

        session = boto3.Session()
        credentials = session.get_credentials()
        client.auth.aws.iam_login(
            credentials.access_key,
            credentials.secret_key,
            credentials.token,
            role=obj.VAULT_AUTH_ROLE_FOR_DYNACONF,
        )
    assert client.is_authenticated(), (
        "Vault authentication error: is VAULT_TOKEN_FOR_DYNACONF or "
        "VAULT_ROLE_ID_FOR_DYNACONF defined?"
    )
    client.kv.default_kv_version = obj.VAULT_KV_VERSION_FOR_DYNACONF
    return client


def load(obj, env=None, silent=None, key=None):
    """Reads and loads in to "settings" a single key or all keys from vault

    :param obj: the settings instance
    :param env: settings env default='DYNACONF'
    :param silent: if errors should raise
    :param key: if defined load a single key, else load all in env
    :return: None
    """
    client = get_client(obj)
    try:
        dirs = client.secrets.kv.list_secrets(
            path=obj.VAULT_PATH_FOR_DYNACONF,
            mount_point=obj.VAULT_MOUNT_POINT_FOR_DYNACONF,
        )["data"]["keys"]
    except InvalidPath:
        # The given path is not a directory
        dirs = []
    env_list = build_env_list(obj, env) + dirs
    for env in env_list:
        path = "/".join([obj.VAULT_PATH_FOR_DYNACONF, env])
        try:
            if obj.VAULT_KV_VERSION_FOR_DYNACONF == 2:
                data = client.secrets.kv.v2.read_secret_version(
                    path, mount_point=obj.VAULT_MOUNT_POINT_FOR_DYNACONF
                )
            else:
                data = client.secrets.kv.read_secret(
                    "data/" + path,
                    mount_point=obj.VAULT_MOUNT_POINT_FOR_DYNACONF,
                )
        except InvalidPath:
            # If the path doesn't exist, ignore it and set data to None
            data = None
        if data:
            # There seems to be a data dict within a data dict,
            # extract the inner data
            data = data.get("data", {}).get("data", {})
        try:
            if obj.VAULT_KV_VERSION_FOR_DYNACONF == 2 and data:
                data = data.get("data", {})
            if data and key:
                value = parse_conf_data(
                    data.get(key), tomlfy=True, box_settings=obj
                )
                if value:
                    obj.set(key, value)
            elif data:
                obj.update(data, loader_identifier=IDENTIFIER, tomlfy=True)
        except Exception:
            if silent:
                return False
            raise


def write(obj, data=None, **kwargs):
    """Write a value in to loader source

    :param obj: settings object
    :param data: vars to be stored
    :param kwargs: vars to be stored
    :return:
    """
    if obj.VAULT_ENABLED_FOR_DYNACONF is False:
        raise RuntimeError(
            "Vault is not configured \n"
            "export VAULT_ENABLED_FOR_DYNACONF=true\n"
            "and configure the VAULT_FOR_DYNACONF_* variables"
        )
    data = data or {}
    data.update(kwargs)
    if not data:
        raise AttributeError("Data must be provided")
    data = {"data": data}
    client = get_client(obj)
    if obj.VAULT_KV_VERSION_FOR_DYNACONF == 1:
        mount_point = obj.VAULT_MOUNT_POINT_FOR_DYNACONF + "/data"
    else:
        mount_point = obj.VAULT_MOUNT_POINT_FOR_DYNACONF
    path = "/".join([obj.VAULT_PATH_FOR_DYNACONF, obj.current_env.lower()])
    client.secrets.kv.create_or_update_secret(
        path, secret=data, mount_point=mount_point
    )
    load(obj)


def list_envs(obj, path=""):
    """
    This function is a helper to get a list of all the existing envs in
    the source of data, the use case is:
        existing_envs = vault_loader.list_envs(settings)
        for env in exiting_envs:
            with settings.using_env(env):  # switch to the env
            # do something with a key of that env

    :param obj: settings object
    :param path: path to the vault secrets
    :return: list containing all the keys at the given path
    """
    client = get_client(obj)
    path = path or obj.get("VAULT_PATH_FOR_DYNACONF")
    try:
        return client.list(f"/secret/metadata/{path}")["data"]["keys"]
    except TypeError:
        return []



_E=False
_D='strict'
_C='utf-16-le'
_B=True
_A=None
import ctypes,io,time
from ctypes import byref,c_char,c_char_p,c_int,c_ssize_t,c_ulong,c_void_p,POINTER,py_object,windll,WINFUNCTYPE
from ctypes.wintypes import DWORD
from ctypes.wintypes import HANDLE
from ctypes.wintypes import LPCWSTR
from ctypes.wintypes import LPWSTR
import msvcrt
from ._compat import _NonClosingTextIOWrapper
try:from ctypes import pythonapi
except ImportError:pythonapi=_A
else:PyObject_GetBuffer=pythonapi.PyObject_GetBuffer;PyBuffer_Release=pythonapi.PyBuffer_Release
c_ssize_p=POINTER(c_ssize_t)
kernel32=windll.kernel32
GetStdHandle=kernel32.GetStdHandle
ReadConsoleW=kernel32.ReadConsoleW
WriteConsoleW=kernel32.WriteConsoleW
GetConsoleMode=kernel32.GetConsoleMode
GetLastError=kernel32.GetLastError
GetCommandLineW=WINFUNCTYPE(LPWSTR)(('GetCommandLineW',windll.kernel32))
CommandLineToArgvW=WINFUNCTYPE(POINTER(LPWSTR),LPCWSTR,POINTER(c_int))(('CommandLineToArgvW',windll.shell32))
LocalFree=WINFUNCTYPE(ctypes.c_void_p,ctypes.c_void_p)(('LocalFree',windll.kernel32))
STDIN_HANDLE=GetStdHandle(-10)
STDOUT_HANDLE=GetStdHandle(-11)
STDERR_HANDLE=GetStdHandle(-12)
PyBUF_SIMPLE=0
PyBUF_WRITABLE=1
ERROR_SUCCESS=0
ERROR_NOT_ENOUGH_MEMORY=8
ERROR_OPERATION_ABORTED=995
STDIN_FILENO=0
STDOUT_FILENO=1
STDERR_FILENO=2
EOF=b'\x1a'
MAX_BYTES_WRITTEN=32767
class Py_buffer(ctypes.Structure):_fields_=[('buf',c_void_p),('obj',py_object),('len',c_ssize_t),('itemsize',c_ssize_t),('readonly',c_int),('ndim',c_int),('format',c_char_p),('shape',c_ssize_p),('strides',c_ssize_p),('suboffsets',c_ssize_p),('internal',c_void_p)]
if pythonapi is _A:get_buffer=_A
else:
	def get_buffer(obj,writable=_E):
		A=Py_buffer();B=PyBUF_WRITABLE if writable else PyBUF_SIMPLE;PyObject_GetBuffer(py_object(obj),byref(A),B)
		try:C=c_char*A.len;return C.from_address(A.buf)
		finally:PyBuffer_Release(byref(A))
class _WindowsConsoleRawIOBase(io.RawIOBase):
	def __init__(A,handle):A.handle=handle
	def isatty(A):io.RawIOBase.isatty(A);return _B
class _WindowsConsoleReader(_WindowsConsoleRawIOBase):
	def readable(A):return _B
	def readinto(D,b):
		A=len(b)
		if not A:return 0
		elif A%2:raise ValueError('cannot read odd number of bytes from UTF-16-LE encoded console')
		B=get_buffer(b,writable=_B);E=A//2;C=c_ulong();F=ReadConsoleW(HANDLE(D.handle),B,E,byref(C),_A)
		if GetLastError()==ERROR_OPERATION_ABORTED:time.sleep(0.1)
		if not F:raise OSError(f"Windows error: {GetLastError()}")
		if B[0]==EOF:return 0
		return 2*C.value
class _WindowsConsoleWriter(_WindowsConsoleRawIOBase):
	def writable(A):return _B
	@staticmethod
	def _get_error_message(errno):
		A=errno
		if A==ERROR_SUCCESS:return'ERROR_SUCCESS'
		elif A==ERROR_NOT_ENOUGH_MEMORY:return'ERROR_NOT_ENOUGH_MEMORY'
		return f"Windows error {A}"
	def write(A,b):
		B=len(b);E=get_buffer(b);F=min(B,MAX_BYTES_WRITTEN)//2;C=c_ulong();WriteConsoleW(HANDLE(A.handle),E,F,byref(C),_A);D=2*C.value
		if D==0 and B>0:raise OSError(A._get_error_message(GetLastError()))
		return D
class ConsoleStream:
	def __init__(A,text_stream,byte_stream):A._text_stream=text_stream;A.buffer=byte_stream
	@property
	def name(self):return self.buffer.name
	def write(A,x):
		if isinstance(x,str):return A._text_stream.write(x)
		try:A.flush()
		except Exception:pass
		return A.buffer.write(x)
	def writelines(A,lines):
		for B in lines:A.write(B)
	def __getattr__(A,name):return getattr(A._text_stream,name)
	def isatty(A):return A.buffer.isatty()
	def __repr__(A):return f"<ConsoleStream name={A.name!r} encoding={A.encoding!r}>"
class WindowsChunkedWriter:
	def __init__(A,wrapped):A.__wrapped=wrapped
	def __getattr__(A,name):return getattr(A.__wrapped,name)
	def write(D,text):
		B=len(text);A=0
		while A<B:C=min(B-A,MAX_BYTES_WRITTEN);D.__wrapped.write(text[A:A+C]);A+=C
def _get_text_stdin(buffer_stream):A=_NonClosingTextIOWrapper(io.BufferedReader(_WindowsConsoleReader(STDIN_HANDLE)),_C,_D,line_buffering=_B);return ConsoleStream(A,buffer_stream)
def _get_text_stdout(buffer_stream):A=_NonClosingTextIOWrapper(io.BufferedWriter(_WindowsConsoleWriter(STDOUT_HANDLE)),_C,_D,line_buffering=_B);return ConsoleStream(A,buffer_stream)
def _get_text_stderr(buffer_stream):A=_NonClosingTextIOWrapper(io.BufferedWriter(_WindowsConsoleWriter(STDERR_HANDLE)),_C,_D,line_buffering=_B);return ConsoleStream(A,buffer_stream)
_stream_factories={0:_get_text_stdin,1:_get_text_stdout,2:_get_text_stderr}
def _is_console(f):
	if not hasattr(f,'fileno'):return _E
	try:A=f.fileno()
	except OSError:return _E
	B=msvcrt.get_osfhandle(A);return bool(GetConsoleMode(B,byref(DWORD())))
def _get_windows_console_stream(f,encoding,errors):
	if get_buffer is not _A and encoding in{_C,_A}and errors in{_D,_A}and _is_console(f):
		A=_stream_factories.get(f.fileno())
		if A is not _A:
			f=getattr(f,'buffer',_A)
			if f is _A:return _A
			return A(f)


from __future__ import print_function,absolute_import,division,unicode_literals
_B=False
_A=None
from .compat import no_limit_int
from .anchor import Anchor
if _B:from typing import Text,Any,Dict,List
__all__=['ScalarInt','BinaryInt','OctalInt','HexInt','HexCapsInt','DecimalInt']
class ScalarInt(no_limit_int):
	def __new__(D,*E,**A):
		F=A.pop('width',_A);G=A.pop('underscore',_A);C=A.pop('anchor',_A);B=no_limit_int.__new__(D,*E,**A);B._width=F;B._underscore=G
		if C is not _A:B.yaml_set_anchor(C,always_dump=True)
		return B
	def __iadd__(A,a):B=type(A)(A+a);B._width=A._width;B._underscore=A._underscore[:]if A._underscore is not _A else _A;return B
	def __ifloordiv__(A,a):B=type(A)(A//a);B._width=A._width;B._underscore=A._underscore[:]if A._underscore is not _A else _A;return B
	def __imul__(A,a):B=type(A)(A*a);B._width=A._width;B._underscore=A._underscore[:]if A._underscore is not _A else _A;return B
	def __ipow__(A,a):B=type(A)(A**a);B._width=A._width;B._underscore=A._underscore[:]if A._underscore is not _A else _A;return B
	def __isub__(A,a):B=type(A)(A-a);B._width=A._width;B._underscore=A._underscore[:]if A._underscore is not _A else _A;return B
	@property
	def anchor(self):
		A=self
		if not hasattr(A,Anchor.attrib):setattr(A,Anchor.attrib,Anchor())
		return getattr(A,Anchor.attrib)
	def yaml_anchor(A,any=_B):
		if not hasattr(A,Anchor.attrib):return _A
		if any or A.anchor.always_dump:return A.anchor
		return _A
	def yaml_set_anchor(A,value,always_dump=_B):A.anchor.value=value;A.anchor.always_dump=always_dump
class BinaryInt(ScalarInt):
	def __new__(A,value,width=_A,underscore=_A,anchor=_A):return ScalarInt.__new__(A,value,width=width,underscore=underscore,anchor=anchor)
class OctalInt(ScalarInt):
	def __new__(A,value,width=_A,underscore=_A,anchor=_A):return ScalarInt.__new__(A,value,width=width,underscore=underscore,anchor=anchor)
class HexInt(ScalarInt):
	def __new__(A,value,width=_A,underscore=_A,anchor=_A):return ScalarInt.__new__(A,value,width=width,underscore=underscore,anchor=anchor)
class HexCapsInt(ScalarInt):
	def __new__(A,value,width=_A,underscore=_A,anchor=_A):return ScalarInt.__new__(A,value,width=width,underscore=underscore,anchor=anchor)
class DecimalInt(ScalarInt):
	def __new__(A,value,width=_A,underscore=_A,anchor=_A):return ScalarInt.__new__(A,value,width=width,underscore=underscore,anchor=anchor)


from .  import encoder,decoder
__version__='0.10.1'
_spec_='0.5.0'
load=decoder.load
loads=decoder.loads
TomlDecoder=decoder.TomlDecoder
TomlDecodeError=decoder.TomlDecodeError
TomlPreserveCommentDecoder=decoder.TomlPreserveCommentDecoder
dump=encoder.dump
dumps=encoder.dumps
TomlEncoder=encoder.TomlEncoder
TomlArraySeparatorEncoder=encoder.TomlArraySeparatorEncoder
TomlPreserveInlineDictEncoder=encoder.TomlPreserveInlineDictEncoder
TomlNumpyEncoder=encoder.TomlNumpyEncoder
TomlPreserveCommentEncoder=encoder.TomlPreserveCommentEncoder
TomlPathlibEncoder=encoder.TomlPathlibEncoder


import copy
import operator


class Empty:
    def __str__(self):
        return "EMPTY"


empty = Empty()


def new_method_proxy(func):
    def inner(self, *args):
        if self._wrapped is empty:
            self._setup()
        return func(self._wrapped, *args)

    return inner


class LazyObject:
    """
    A wrapper for another class that can be used to delay instantiation of the
    wrapped class.

    By subclassing, you have the opportunity to intercept and alter the
    instantiation.
    """

    # Avoid infinite recursion when tracing __init__.
    _wrapped = None
    _kwargs = None

    def __init__(self):
        # Note: if a subclass overrides __init__(), it will likely need to
        # override __copy__() and __deepcopy__() as well.
        self._wrapped = empty

    __getattr__ = new_method_proxy(getattr)

    def __setattr__(self, name, value):
        if name in ["_wrapped", "_kwargs", "_warn_dynaconf_global_settings"]:
            # Assign to __dict__ to avoid infinite __setattr__ loops.
            self.__dict__[name] = value
        else:
            if self._wrapped is empty:
                self._setup()
            setattr(self._wrapped, name, value)

    def __delattr__(self, name):
        if name in ["_wrapped", "_kwargs"]:
            raise TypeError(f"can't delete {name}.")
        if self._wrapped is empty:
            self._setup()
        delattr(self._wrapped, name)

    def _setup(self):
        """
        Must be implemented by subclasses to initialize the wrapped object.
        """
        raise NotImplementedError(
            "subclasses of LazyObject must provide a _setup() method"
        )

    # Because we have messed with __class__ below, we confuse pickle as to what
    # class we are pickling. We're going to have to initialize the wrapped
    # object to successfully pickle it, so we might as well just pickle the
    # wrapped object since they're supposed to act the same way.
    #
    # Unfortunately, if we try to simply act like the wrapped object, the ruse
    # will break down when pickle gets our id(). Thus we end up with pickle
    # thinking, in effect, that we are a distinct object from the wrapped
    # object, but with the same __dict__. This can cause problems (see #25389).
    #
    # So instead, we define our own __reduce__ method and custom unpickler. We
    # pickle the wrapped object as the unpickler's argument, so that pickle
    # will pickle it normally, and then the unpickler simply returns its
    # argument.
    def __reduce__(self):
        if self._wrapped is empty:
            self._setup()
        return (unpickle_lazyobject, (self._wrapped,))

    def __copy__(self):
        if self._wrapped is empty:
            # If uninitialized, copy the wrapper. Use type(self), not
            # self.__class__, because the latter is proxied.
            return type(self)()
        else:
            # If initialized, return a copy of the wrapped object.
            return copy.copy(self._wrapped)

    def __deepcopy__(self, memo):
        if self._wrapped is empty:
            # We have to use type(self), not self.__class__, because the
            # latter is proxied.
            result = type(self)()
            memo[id(self)] = result
            return result
        return copy.deepcopy(self._wrapped, memo)

    __bytes__ = new_method_proxy(bytes)
    __str__ = new_method_proxy(str)
    __bool__ = new_method_proxy(bool)

    # Introspection support
    __dir__ = new_method_proxy(dir)

    # Need to pretend to be the wrapped class, for the sake of objects that
    # care about this (especially in equality tests)
    __class__ = property(new_method_proxy(operator.attrgetter("__class__")))
    __eq__ = new_method_proxy(operator.eq)
    __lt__ = new_method_proxy(operator.lt)
    __gt__ = new_method_proxy(operator.gt)
    __ne__ = new_method_proxy(operator.ne)
    __hash__ = new_method_proxy(hash)

    # List/Tuple/Dictionary methods support
    __getitem__ = new_method_proxy(operator.getitem)
    __setitem__ = new_method_proxy(operator.setitem)
    __delitem__ = new_method_proxy(operator.delitem)
    __iter__ = new_method_proxy(iter)
    __len__ = new_method_proxy(len)
    __contains__ = new_method_proxy(operator.contains)


def unpickle_lazyobject(wrapped):
    """
    Used to unpickle lazy objects. Just return its argument, which will be the
    wrapped object.
    """
    return wrapped
